from pathlib import Path

import numpy as np
import pytest
import torch
from scipy.stats import wasserstein_distance

from careamics.config import GaussianMixtureNMConfig, MultiChannelNMConfig
from careamics.models.lvae.likelihoods import NoiseModelLikelihood
from careamics.models.lvae.noise_models import (
    GaussianMixtureNoiseModel,
    MultiChannelNoiseModel,
    noise_model_factory,
)

pytestmark = pytest.mark.lvae


def test_factory_no_noise_model():
    noise_model = noise_model_factory(None)
    assert noise_model is None


def test_instantiate_noise_model(tmp_path: Path, create_dummy_noise_model) -> None:
    # Create a dummy noise model
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)

    # Instantiate the noise model
    gmm = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        path=tmp_path / "dummy_noise_model.npz",
        # all other params are default
    )
    noise_model_config = MultiChannelNMConfig(noise_models=[gmm])
    noise_model = noise_model_factory(noise_model_config)
    assert noise_model is not None
    assert noise_model.nmodel_0.weight.shape == (9, 3)
    assert noise_model.nmodel_0.min_signal == 0
    assert noise_model.nmodel_0.max_signal == 2**16 - 1
    assert noise_model.nmodel_0.min_sigma == 0.125


def test_instantiate_multiple_noise_models(
    tmp_path: Path, create_dummy_noise_model
) -> None:
    # Create a dummy noise model
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)

    # Instantiate the noise model
    gmm = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        path=tmp_path / "dummy_noise_model.npz",
        # all other params are default
    )
    noise_model_config = MultiChannelNMConfig(noise_models=[gmm, gmm, gmm])
    noise_model = noise_model_factory(noise_model_config)
    assert noise_model is not None
    assert noise_model.nmodel_0 is not None
    assert noise_model.nmodel_1 is not None
    assert noise_model.nmodel_2 is not None
    assert noise_model.nmodel_0.weight.shape == (9, 3)
    assert noise_model.nmodel_0.min_signal == 0
    assert noise_model.nmodel_0.max_signal == 2**16 - 1
    assert noise_model.nmodel_0.min_sigma == 0.125
    assert noise_model.nmodel_1.weight.shape == (9, 3)
    assert noise_model.nmodel_1.min_signal == 0
    assert noise_model.nmodel_1.max_signal == 2**16 - 1
    assert noise_model.nmodel_1.min_sigma == 0.125
    assert noise_model.nmodel_2.weight.shape == (9, 3)
    assert noise_model.nmodel_2.min_signal == 0
    assert noise_model.nmodel_2.max_signal == 2**16 - 1
    assert noise_model.nmodel_2.min_sigma == 0.125


@pytest.mark.parametrize("img_size", [64, 128])
def test_noise_model_likelihood(
    tmp_path: Path,
    img_size: int,
    create_dummy_noise_model,
) -> None:
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)

    gmm_config = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        path=tmp_path / "dummy_noise_model.npz",
        # all other params are default
    )
    nm = GaussianMixtureNoiseModel(gmm_config)
    assert nm is not None
    assert isinstance(nm, GaussianMixtureNoiseModel)

    inp_shape = (1, 1, img_size, img_size)
    signal = torch.ones(inp_shape)
    obs = signal + torch.randn(inp_shape) * 0.1
    likelihood = nm.likelihood(obs, signal)
    assert likelihood.shape == inp_shape


@pytest.mark.parametrize("img_size", [64, 128])
@pytest.mark.parametrize("target_ch", [1, 2, 3])
def test_multi_channel_noise_model_likelihood(
    tmp_path: Path,
    img_size: int,
    target_ch: int,
    create_dummy_noise_model,
) -> None:
    noise_models = []
    rand_epss = []
    for i in range(target_ch):
        eps = np.random.rand()
        nm_dict = create_dummy_noise_model.copy()
        nm_dict["trained_weight"] = nm_dict["trained_weight"] + eps
        rand_epss.append(eps)
        np.savez(tmp_path / f"dummy_noise_model_{i}.npz", **nm_dict)

        gmm = GaussianMixtureNMConfig(
            model_type="GaussianMixtureNoiseModel",
            path=tmp_path / f"dummy_noise_model_{i}.npz",
            # all other params are default
        )
        noise_models.append(gmm)

    noise_model_config = MultiChannelNMConfig(noise_models=noise_models)
    nm = noise_model_factory(noise_model_config)
    assert nm is not None
    assert isinstance(nm, MultiChannelNoiseModel)
    assert nm._nm_cnt == target_ch
    assert all(
        isinstance(getattr(nm, f"nmodel_{i}"), GaussianMixtureNoiseModel)
        for i in range(nm._nm_cnt)
    )
    assert all(
        np.allclose(
            getattr(nm, f"nmodel_{i}").weight,
            create_dummy_noise_model["trained_weight"] + rand_epss[i],
        )
        for i in range(nm._nm_cnt)
    )
    inp_shape = (1, target_ch, img_size, img_size)
    signal = torch.ones(inp_shape)
    obs = signal + torch.randn(inp_shape) * 0.1
    likelihood = nm.likelihood(obs, signal)
    assert likelihood.shape == inp_shape


@pytest.mark.parametrize(
    "image_size, max_value, noise_scale",
    [
        ([5, 128, 128], 255, 0.1),
        ([5, 128, 128], 255, 0.5),
    ],
)
def test_gm_noise_model_training(image_size, max_value, noise_scale):
    gen = np.random.default_rng(42)
    signal_normalized = gen.uniform(0, 1, image_size)
    noise = gen.normal(0, noise_scale, image_size)
    observation_normalized = signal_normalized + noise
    signal = signal_normalized * max_value
    observation = observation_normalized * max_value

    nm_config = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        n_gaussian=1,
        min_signal=signal.min(),
        max_signal=signal.max(),
    )
    noise_model = GaussianMixtureNoiseModel(nm_config)
    training_losses = noise_model.fit(
        signal=signal, observation=observation, n_epochs=500
    )
    initial_loss = training_losses[0]
    last_loss = training_losses[-1]
    # Check if model is training
    assert initial_loss > last_loss

    # check if estimated mean and std of a noisy sample are close to real ones
    signal_tensor = torch.from_numpy(signal).to(torch.float32)
    mus, sigmas, _ = noise_model.get_gaussian_parameters(signal_tensor)

    # learned mean should be close to the mean of the signal
    learned_mu = mus.mean() / max_value
    real_mu = signal_normalized.mean()
    assert np.allclose(learned_mu, real_mu, atol=1e-2)

    # learned sigma should be close to the noise sigma
    learned_sigma = sigmas.mean() / max_value
    noise_image = observation_normalized - signal_normalized
    real_sigma = noise_image.std()
    assert np.allclose(learned_sigma, real_sigma, atol=1e-2)


@pytest.mark.parametrize("image_size, max_value", [([256, 256], 255)])
def test_noise_model_sampling(image_size, max_value):
    gen = np.random.default_rng(42)

    signal = gen.uniform(0, 1, image_size)
    observation = signal + gen.normal(0, 0.1, signal.shape)
    signal = signal * max_value
    observation = observation * max_value

    nm_config = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel",
        n_gaussian=1,
        min_sigma=100,
        min_signal=signal.min(),
        max_signal=signal.max(),
    )
    noise_model = GaussianMixtureNoiseModel(nm_config)
    noise_model.fit(signal=signal, observation=observation, n_epochs=200)
    sampled_noise_data = noise_model.sample_observation_from_signal(signal)
    assert sampled_noise_data.shape == signal.shape

    real_noise = observation - signal
    predicted_noise = sampled_noise_data - signal
    real_noise = real_noise / max_value
    predicted_noise = predicted_noise / max_value
    noise_distribution_difference = wasserstein_distance(
        real_noise.ravel(), predicted_noise.ravel()
    )
    assert noise_distribution_difference < 0.1


def test_noise_model_in_likelihood_call():
    test_input = torch.rand(256, 256)
    test_target = torch.rand(256, 256)

    nm_config = GaussianMixtureNMConfig(
        model_type="GaussianMixtureNoiseModel", n_gaussian=1
    )
    noise_model = GaussianMixtureNoiseModel(nm_config)
    likelihood = NoiseModelLikelihood(
        data_mean=test_input.mean(), data_std=test_input.std(), noise_model=noise_model
    )
    log_likelihood, _ = likelihood(test_input, test_target)
    assert log_likelihood is not None



import pytest

from careamics.config.algorithms import N2VAlgorithm
from careamics.config.support import SupportedPixelManipulation


def test_n_channels_n2v():
    """Check that an error is raised if n2v has different number of channels in
    input and output."""
    model = {
        "architecture": "UNet",
        "in_channels": 1,
        "num_classes": 2,
        "n2v2": False,
    }
    loss = "n2v"

    with pytest.raises(ValueError):
        N2VAlgorithm(algorithm="n2v", loss=loss, model=model)


def test_channels(minimum_algorithm_n2v: dict):
    """Check that error is thrown if the number of channels are different."""
    minimum_algorithm_n2v["model"] = {
        "architecture": "UNet",
        "in_channels": 2,
        "num_classes": 2,
        "n2v2": False,
    }
    N2VAlgorithm(**minimum_algorithm_n2v)

    minimum_algorithm_n2v["model"]["num_classes"] = 3
    with pytest.raises(ValueError):
        N2VAlgorithm(**minimum_algorithm_n2v)


def test_no_final_activation(minimum_algorithm_n2v: dict):
    """Check that an error is raised if the model has a final activation."""
    minimum_algorithm_n2v["model"] = {
        "architecture": "UNet",
        "final_activation": "ReLU",
    }
    with pytest.raises(ValueError):
        N2VAlgorithm(**minimum_algorithm_n2v)


def test_set_n2v_strategy(minimum_algorithm_n2v: dict):
    """Test that the N2V2 can be set."""
    uniform = SupportedPixelManipulation.UNIFORM.value
    median = SupportedPixelManipulation.MEDIAN.value

    # Test default
    cfg = N2VAlgorithm(**minimum_algorithm_n2v)
    assert cfg.n2v_config.strategy == uniform

    # Test setting to n2v2
    cfg.set_n2v2(True)
    assert cfg.n2v_config.strategy == median
    assert cfg.model.n2v2 is True

    # Test setting back to n2v
    cfg.set_n2v2(False)
    assert cfg.n2v_config.strategy == uniform
    assert cfg.model.n2v2 is False



from pathlib import Path

import numpy as np
import pytest
import tifffile

from careamics.config import InferenceConfig, create_n2n_configuration
from careamics.dataset_ng.dataset.dataset import CareamicsDataset, Mode


# TODO: these tests not deterministic anymore - DataConfig doesn't have random seed
# could make a mock? temporarily
@pytest.mark.parametrize(
    "data_shape, patch_size, expected_dataset_len",
    [
        ((256, 256), (32, 32), 64),
        ((512, 512), (64, 64), 64),
        ((128, 128), (32, 32), 16),
    ],
)
def test_from_array(data_shape, patch_size, expected_dataset_len):
    rng = np.random.default_rng(42)
    example_input = rng.random(data_shape)
    example_target = rng.random(data_shape)

    train_data_config = create_n2n_configuration(
        "test_exp",
        data_type="array",
        axes="YX",
        patch_size=patch_size,
        batch_size=1,
        num_epochs=1,
    ).data_config

    train_dataset = CareamicsDataset(
        data_config=train_data_config,
        mode=Mode.TRAINING,
        inputs=[example_input],
        targets=[example_target],
    )

    assert len(train_dataset) == expected_dataset_len
    sample, target = train_dataset[0]
    assert sample.data.shape == (1, *patch_size)
    assert target.data.shape == (1, *patch_size)


@pytest.mark.parametrize(
    "data_shape, patch_size, expected_dataset_len",
    [
        ((256, 256), (32, 32), 64),
        ((512, 512), (64, 64), 64),
        ((128, 128), (32, 32), 16),
    ],
)
def test_from_tiff(tmp_path: Path, data_shape, patch_size, expected_dataset_len):
    rng = np.random.default_rng(42)
    example_input = rng.random(data_shape)
    example_target = rng.random(data_shape)

    input_file_path = tmp_path / "input.tiff"
    target_file_path = tmp_path / "target.tiff"

    tifffile.imwrite(input_file_path, example_input)
    tifffile.imwrite(target_file_path, example_target)

    train_data_config = create_n2n_configuration(
        "test_exp",
        data_type="tiff",
        axes="YX",
        patch_size=patch_size,
        batch_size=1,
        num_epochs=1,
    ).data_config

    train_dataset = CareamicsDataset(
        data_config=train_data_config,
        mode=Mode.TRAINING,
        inputs=[input_file_path],
        targets=[target_file_path],
    )

    assert len(train_dataset) == expected_dataset_len
    sample, target = train_dataset[0]
    assert sample.data.shape == (1, *patch_size)
    assert target.data.shape == (1, *patch_size)


@pytest.mark.skip("Prediction not fully implemented")
@pytest.mark.parametrize(
    "data_shape, tile_size, tile_overlap",
    [
        ((256, 256), (32, 32), (16, 16)),
        ((512, 512), (64, 64), (32, 32)),
        ((128, 128), (32, 32), (8, 8)),
    ],
)
def test_prediction_from_array(data_shape, tile_size, tile_overlap):
    rng = np.random.default_rng(42)
    example_data = rng.random(data_shape)

    prediction_config = InferenceConfig(
        data_type="array",
        tile_size=tile_size,
        tile_overlap=tile_overlap,
        axes="YX",
        image_means=(example_data.mean(),),
        image_stds=(example_data.std(),),
        tta_transforms=False,
        batch_size=1,
    )

    prediction_dataset = CareamicsDataset(
        data_config=prediction_config,
        mode=Mode.PREDICTING,
        inputs=[example_data],
    )

    assert len(prediction_dataset) > 0
    sample, target = prediction_dataset[0]
    assert sample.data.shape == (1, *tile_size)
    assert target is None


@pytest.mark.parametrize(
    "patch_size, data_shape",
    [
        ((32, 32), (256, 256)),
        ((64, 64), (512, 512)),
        ((16, 16), (128, 128)),
    ],
)
def test_from_custom_data_type(patch_size, data_shape):
    rng = np.random.default_rng(42)
    example_data = rng.random(data_shape)
    example_target = rng.random(data_shape)

    train_data_config = create_n2n_configuration(
        "test_exp",
        data_type="custom",
        axes="YX",
        patch_size=patch_size,
        batch_size=1,
        num_epochs=1,
    ).data_config

    def read_data_func_test(data):
        return 1 - data

    train_dataset = CareamicsDataset(
        data_config=train_data_config,
        mode=Mode.TRAINING,
        inputs=[example_data],
        targets=[example_target],
        read_func=read_data_func_test,
        read_kwargs={},
    )

    assert len(train_dataset) > 0
    sample, target = train_dataset[0]
    assert sample.data.shape == (1, *patch_size)
    assert target.data.shape == (1, *patch_size)



"""Iterable dataset used to load data file by file."""

from __future__ import annotations

import copy
from collections.abc import Generator
from pathlib import Path
from typing import Callable, Optional

import numpy as np
from torch.utils.data import IterableDataset

from careamics.config import DataConfig
from careamics.config.transformations import NormalizeModel
from careamics.file_io.read import read_tiff
from careamics.transforms import Compose

from ..utils.logging import get_logger
from .dataset_utils import iterate_over_files
from .dataset_utils.running_stats import WelfordStatistics
from .patching.patching import Stats
from .patching.random_patching import extract_patches_random

logger = get_logger(__name__)


class PathIterableDataset(IterableDataset):
    """
    Dataset allowing extracting patches w/o loading whole data into memory.

    Parameters
    ----------
    data_config : DataConfig
        Data configuration.
    src_files : list of pathlib.Path
        List of data files.
    target_files : list of pathlib.Path, optional
        Optional list of target files, by default None.
    read_source_func : Callable, optional
        Read source function for custom types, by default read_tiff.

    Attributes
    ----------
    data_path : list of pathlib.Path
        Path to the data, must be a directory.
    axes : str
        Description of axes in format STCZYX.
    """

    def __init__(
        self,
        data_config: DataConfig,
        src_files: list[Path],
        target_files: Optional[list[Path]] = None,
        read_source_func: Callable = read_tiff,
    ) -> None:
        """Constructors.

        Parameters
        ----------
        data_config : GeneralDataConfig
            Data configuration.
        src_files : list[Path]
            List of data files.
        target_files : list[Path] or None, optional
            Optional list of target files, by default None.
        read_source_func : Callable, optional
            Read source function for custom types, by default read_tiff.
        """
        self.data_config = data_config
        self.data_files = src_files
        self.target_files = target_files
        self.read_source_func = read_source_func

        # compute mean and std over the dataset
        # only checking the image_mean because the DataConfig class ensures that
        # if image_mean is provided, image_std is also provided
        if not self.data_config.image_means:
            self.image_stats, self.target_stats = self._calculate_mean_and_std()
            logger.info(
                f"Computed dataset mean: {self.image_stats.means},"
                f"std: {self.image_stats.stds}"
            )

            # update the mean in the config
            self.data_config.set_means_and_stds(
                image_means=self.image_stats.means,
                image_stds=self.image_stats.stds,
                target_means=(
                    list(self.target_stats.means)
                    if self.target_stats.means is not None
                    else None
                ),
                target_stds=(
                    list(self.target_stats.stds)
                    if self.target_stats.stds is not None
                    else None
                ),
            )

        else:
            # if mean and std are provided in the config, use them
            self.image_stats, self.target_stats = (
                Stats(self.data_config.image_means, self.data_config.image_stds),
                Stats(self.data_config.target_means, self.data_config.target_stds),
            )

        # create transform composed of normalization and other transforms
        self.patch_transform = Compose(
            transform_list=[
                NormalizeModel(
                    image_means=self.image_stats.means,
                    image_stds=self.image_stats.stds,
                    target_means=self.target_stats.means,
                    target_stds=self.target_stats.stds,
                )
            ]
            + list(data_config.transforms)
        )

    def _calculate_mean_and_std(self) -> tuple[Stats, Stats]:
        """
        Calculate mean and std of the dataset.

        Returns
        -------
        tuple of Stats and optional Stats
            Data classes containing the image and target statistics.
        """
        num_samples = 0
        image_stats = WelfordStatistics()
        if self.target_files is not None:
            target_stats = WelfordStatistics()

        for sample, target in iterate_over_files(
            self.data_config, self.data_files, self.target_files, self.read_source_func
        ):
            # update the image statistics
            image_stats.update(sample, num_samples)

            # update the target statistics if target is available
            if target is not None:
                target_stats.update(target, num_samples)

            num_samples += 1

        if num_samples == 0:
            raise ValueError("No samples found in the dataset.")

        # Average the means and stds per sample
        image_means, image_stds = image_stats.finalize()

        if target is not None:
            target_means, target_stds = target_stats.finalize()

            return (
                Stats(image_means, image_stds),
                Stats(np.array(target_means), np.array(target_stds)),
            )
        else:
            return Stats(image_means, image_stds), Stats(None, None)

    def __iter__(
        self,
    ) -> Generator[tuple[np.ndarray, ...], None, None]:
        """
        Iterate over data source and yield single patch.

        Yields
        ------
        np.ndarray
            Single patch.
        """
        assert (
            self.image_stats.means is not None and self.image_stats.stds is not None
        ), "Mean and std must be provided"

        # iterate over files
        for sample_input, sample_target in iterate_over_files(
            self.data_config, self.data_files, self.target_files, self.read_source_func
        ):
            patches = extract_patches_random(
                arr=sample_input,
                patch_size=self.data_config.patch_size,
                target=sample_target,
            )

            # iterate over patches
            # patches are tuples of (patch, target) if target is available
            # or (patch, None) only if no target is available
            # patch is of dimensions (C)ZYX
            for patch_data in patches:
                yield self.patch_transform(
                    patch=patch_data[0],
                    target=patch_data[1],
                )

    def get_data_statistics(self) -> tuple[list[float], list[float]]:
        """Return training data statistics.

        Returns
        -------
        tuple of list of floats
            Means and standard deviations across channels of the training data.
        """
        return self.image_stats.get_statistics()

    def get_number_of_files(self) -> int:
        """
        Return the number of files in the dataset.

        Returns
        -------
        int
            Number of files in the dataset.
        """
        return len(self.data_files)

    def split_dataset(
        self,
        percentage: float = 0.1,
        minimum_number: int = 5,
    ) -> PathIterableDataset:
        """Split up dataset in two.

        Splits the datest sing a percentage of the data (files) to extract, or the
        minimum number of the percentage is less than the minimum number.

        Parameters
        ----------
        percentage : float, optional
            Percentage of files to split up, by default 0.1.
        minimum_number : int, optional
            Minimum number of files to split up, by default 5.

        Returns
        -------
        IterableDataset
            Dataset containing the split data.

        Raises
        ------
        ValueError
            If the percentage is smaller than 0 or larger than 1.
        ValueError
            If the minimum number is smaller than 1 or larger than the number of files.
        """
        if percentage < 0 or percentage > 1:
            raise ValueError(f"Percentage must be between 0 and 1, got {percentage}.")

        if minimum_number < 1 or minimum_number > self.get_number_of_files():
            raise ValueError(
                f"Minimum number of files must be between 1 and "
                f"{self.get_number_of_files()} (number of files), got "
                f"{minimum_number}."
            )

        # compute number of files
        total_files = self.get_number_of_files()
        n_files = max(round(percentage * total_files), minimum_number)

        # get random indices
        indices = np.random.choice(total_files, n_files, replace=False)

        # extract files
        val_files = [self.data_files[i] for i in indices]

        # remove patches from self.patch
        data_files = []
        for i, file in enumerate(self.data_files):
            if i not in indices:
                data_files.append(file)
        self.data_files = data_files

        # same for targets
        if self.target_files is not None:
            val_target_files = [self.target_files[i] for i in indices]

            data_target_files = []
            for i, file in enumerate(self.target_files):
                if i not in indices:
                    data_target_files.append(file)
            self.target_files = data_target_files

        # clone the dataset
        dataset = copy.deepcopy(self)

        # reassign patches
        dataset.data_files = val_files

        # reassign targets
        if self.target_files is not None:
            dataset.target_files = val_target_files

        return dataset



"""Pydantic model for the Normalize transform."""

from typing import Literal, Optional

from pydantic import ConfigDict, Field, model_validator
from typing_extensions import Self

from .transform_model import TransformModel


class NormalizeModel(TransformModel):
    """
    Pydantic model used to represent Normalize transformation.

    The Normalize transform is a zero mean and unit variance transformation.

    Attributes
    ----------
    name : Literal["Normalize"]
        Name of the transformation.
    mean : float
        Mean value for normalization.
    std : float
        Standard deviation value for normalization.
    """

    model_config = ConfigDict(
        validate_assignment=True,
    )

    name: Literal["Normalize"] = "Normalize"
    image_means: list = Field(..., min_length=0, max_length=32)
    image_stds: list = Field(..., min_length=0, max_length=32)
    target_means: Optional[list] = Field(default=None, min_length=0, max_length=32)
    target_stds: Optional[list] = Field(default=None, min_length=0, max_length=32)

    @model_validator(mode="after")
    def validate_means_stds(self: Self) -> Self:
        """Validate that the means and stds have the same length.

        Returns
        -------
        Self
            The instance of the model.
        """
        if len(self.image_means) != len(self.image_stds):
            raise ValueError("The number of image means and stds must be the same.")

        if (self.target_means is None) != (self.target_stds is None):
            raise ValueError(
                "Both target means and stds must be provided together, or bot None."
            )

        if self.target_means is not None and self.target_stds is not None:
            if len(self.target_means) != len(self.target_stds):
                raise ValueError(
                    "The number of target means and stds must be the same."
                )

        return self



"""Functions relating reading and writing image files."""

__all__ = [
    "ReadFunc",
    "SupportedWriteType",
    "WriteFunc",
    "get_read_func",
    "get_write_func",
    "read",
    "write",
]

from . import read, write
from .read import ReadFunc, get_read_func
from .write import SupportedWriteType, WriteFunc, get_write_func



from collections.abc import Sequence
from enum import Enum
from pathlib import Path
from typing import Literal, NamedTuple, Optional, Union

import numpy as np
from numpy.typing import NDArray
from torch.utils.data import Dataset
from typing_extensions import ParamSpec

from careamics.config import DataConfig, InferenceConfig
from careamics.config.support import SupportedData
from careamics.dataset.patching.patching import Stats
from careamics.dataset_ng.patch_extractor import (
    ImageStackLoader,
    PatchExtractor,
    create_patch_extractor,
)
from careamics.dataset_ng.patching_strategies import (
    FixedRandomPatchingStrategy,
    PatchingStrategy,
    PatchSpecs,
    RandomPatchingStrategy,
)
from careamics.transforms import Compose

P = ParamSpec("P")


class Mode(str, Enum):
    TRAINING = "training"
    VALIDATING = "validating"
    PREDICTING = "predicting"


class ImageRegionData(NamedTuple):
    data: NDArray
    source: Union[Path, Literal["array"]]
    data_shape: Sequence[int]
    dtype: str  # dtype should be str for collate
    axes: str
    region_spec: PatchSpecs


InputType = Union[Sequence[np.ndarray], Sequence[Path]]


class CareamicsDataset(Dataset):
    def __init__(
        self,
        data_config: Union[DataConfig, InferenceConfig],
        mode: Mode,
        inputs: InputType,
        targets: Optional[InputType] = None,
        image_stack_loader: Optional[ImageStackLoader[P]] = None,
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        self.config = data_config
        self.mode = mode

        data_type_enum = SupportedData(self.config.data_type)
        self.input_extractor = create_patch_extractor(
            inputs,
            self.config.axes,
            data_type_enum,
            image_stack_loader,
            *args,
            **kwargs,
        )
        if targets is not None:
            self.target_extractor: Optional[PatchExtractor] = create_patch_extractor(
                targets,
                self.config.axes,
                data_type_enum,
                image_stack_loader,
                *args,
                **kwargs,
            )
        else:
            self.target_extractor = None

        self.patching_strategy = self._initialize_patching_strategy()

        self.input_stats, self.target_stats = self._initialize_statistics()

        self.transforms = self._initialize_transforms()

    def _initialize_patching_strategy(self) -> PatchingStrategy:
        patching_strategy: PatchingStrategy
        if self.mode == Mode.TRAINING:
            if isinstance(self.config, InferenceConfig):
                raise ValueError("Inference config cannot be used for training.")
            patching_strategy = RandomPatchingStrategy(
                data_shapes=self.input_extractor.shape,
                patch_size=self.config.patch_size,
                # TODO: Add random seed to dataconfig
                seed=getattr(self.config, "random_seed", None),
            )
        elif self.mode == Mode.VALIDATING:
            if isinstance(self.config, InferenceConfig):
                raise ValueError("Inference config cannot be used for validating.")
            patching_strategy = FixedRandomPatchingStrategy(
                data_shapes=self.input_extractor.shape,
                patch_size=self.config.patch_size,
                # TODO: Add random seed to dataconfig
                seed=getattr(self.config, "random_seed", None),
            )
        elif self.mode == Mode.PREDICTING:
            # TODO: patching strategy will be tilingStrategy in upcoming PR
            raise NotImplementedError(
                "Prediction mode for the CAREamicsDataset has not been implemented yet."
            )
        else:
            raise ValueError(f"Unrecognised dataset mode {self.mode}.")

        return patching_strategy

    def _initialize_transforms(self) -> Optional[Compose]:
        if isinstance(self.config, DataConfig):
            return Compose(
                transform_list=list(self.config.transforms),
            )
        # TODO: add TTA
        return None

    def _initialize_statistics(self) -> tuple[Stats, Optional[Stats]]:
        # TODO: add running stats
        # Currently assume that stats are provided in the configuration
        input_stats = Stats(self.config.image_means, self.config.image_stds)
        target_stats = None
        if isinstance(self.config, DataConfig):
            target_means = self.config.target_means
            target_stds = self.config.target_stds
            if target_means is not None and target_stds is not None:
                target_stats = Stats(target_means, target_stds)
        return input_stats, target_stats

    def __len__(self):
        return self.patching_strategy.n_patches

    def _create_image_region(
        self, patch: np.ndarray, patch_spec: PatchSpecs, extractor: PatchExtractor
    ) -> ImageRegionData:
        data_idx = patch_spec["data_idx"]
        return ImageRegionData(
            data=patch,
            source=extractor.image_stacks[data_idx].source,
            dtype=str(extractor.image_stacks[data_idx].data_dtype),
            data_shape=extractor.image_stacks[data_idx].data_shape,
            # TODO: should it be axes of the original image instead?
            axes=self.config.axes,
            region_spec=patch_spec,
        )

    def __getitem__(
        self, index: int
    ) -> tuple[ImageRegionData, Optional[ImageRegionData]]:
        patch_spec = self.patching_strategy.get_patch_spec(index)
        input_patch = self.input_extractor.extract_patch(**patch_spec)

        target_patch = (
            self.target_extractor.extract_patch(**patch_spec)
            if self.target_extractor is not None
            else None
        )

        if self.transforms is not None:
            input_patch, target_patch = self.transforms(input_patch, target_patch)

        input_data = self._create_image_region(
            patch=input_patch, patch_spec=patch_spec, extractor=self.input_extractor
        )

        if target_patch is not None and self.target_extractor is not None:
            target_data = self._create_image_region(
                patch=target_patch,
                patch_spec=patch_spec,
                extractor=self.target_extractor,
            )
        else:
            target_data = None

        return input_data, target_data



"""UNet-based algorithm Pydantic model."""

from pprint import pformat
from typing import Literal

from pydantic import BaseModel, ConfigDict

from careamics.config.architectures import UNetModel
from careamics.config.optimizer_models import LrSchedulerModel, OptimizerModel


class UNetBasedAlgorithm(BaseModel):
    """General UNet-based algorithm configuration.

    This Pydantic model validates the parameters governing the components of the
    training algorithm: which algorithm, loss function, model architecture, optimizer,
    and learning rate scheduler to use.

    Currently, we only support N2V, CARE, and N2N algorithms. In order to train these
    algorithms, use the corresponding configuration child classes (e.g.
    `N2VAlgorithm`) to ensure coherent parameters (e.g. specific losses).


    Attributes
    ----------
    algorithm : {"n2v", "care", "n2n"}
        Algorithm to use.
    loss : {"n2v", "mae", "mse"}
        Loss function to use.
    model : UNetModel
        Model architecture to use.
    optimizer : OptimizerModel, optional
        Optimizer to use.
    lr_scheduler : LrSchedulerModel, optional
        Learning rate scheduler to use.

    Raises
    ------
    ValueError
        Algorithm parameter type validation errors.
    ValueError
        If the algorithm, loss and model are not compatible.
    """

    # Pydantic class configuration
    model_config = ConfigDict(
        protected_namespaces=(),  # allows to use model_* as a field name
        validate_assignment=True,
        extra="allow",
    )

    # Mandatory fields
    algorithm: Literal["n2v", "care", "n2n"]
    """Algorithm name, as defined in SupportedAlgorithm."""

    loss: Literal["n2v", "mae", "mse"]
    """Loss function to use, as defined in SupportedLoss."""

    model: UNetModel
    """UNet model configuration."""

    # Optional fields
    optimizer: OptimizerModel = OptimizerModel()
    """Optimizer to use, defined in SupportedOptimizer."""

    lr_scheduler: LrSchedulerModel = LrSchedulerModel()
    """Learning rate scheduler to use, defined in SupportedLrScheduler."""

    def __str__(self) -> str:
        """Pretty string representing the configuration.

        Returns
        -------
        str
            Pretty string.
        """
        return pformat(self.model_dump())

    @classmethod
    def get_compatible_algorithms(cls) -> list[str]:
        """Get the list of compatible algorithms.

        Returns
        -------
        list of str
            List of compatible algorithms.
        """
        return ["n2v", "care", "n2n"]



__all__ = [
    "ImageStackLoader",
    "PatchExtractor",
    "create_patch_extractor",
    "get_image_stack_loader",
]

from .image_stack_loader import ImageStackLoader, get_image_stack_loader
from .patch_extractor import PatchExtractor
from .patch_extractor_factory import create_patch_extractor



"""In-memory prediction dataset."""

from __future__ import annotations

from numpy.typing import NDArray
from torch.utils.data import Dataset

from careamics.transforms import Compose

from ..config import InferenceConfig
from ..config.transformations import NormalizeModel
from .dataset_utils import reshape_array


class InMemoryPredDataset(Dataset):
    """Simple prediction dataset returning images along the sample axis.

    Parameters
    ----------
    prediction_config : InferenceConfig
        Prediction configuration.
    inputs : NDArray
        Input data.
    """

    def __init__(
        self,
        prediction_config: InferenceConfig,
        inputs: NDArray,
    ) -> None:
        """Constructor.

        Parameters
        ----------
        prediction_config : InferenceConfig
            Prediction configuration.
        inputs : NDArray
            Input data.

        Raises
        ------
        ValueError
            If data_path is not a directory.
        """
        self.pred_config = prediction_config
        self.input_array = inputs
        self.axes = self.pred_config.axes
        self.image_means = self.pred_config.image_means
        self.image_stds = self.pred_config.image_stds

        # Reshape data
        self.data = reshape_array(self.input_array, self.axes)

        # get transforms
        self.patch_transform = Compose(
            transform_list=[
                NormalizeModel(image_means=self.image_means, image_stds=self.image_stds)
            ],
        )

    def __len__(self) -> int:
        """
        Return the length of the dataset.

        Returns
        -------
        int
            Length of the dataset.
        """
        return len(self.data)

    def __getitem__(self, index: int) -> tuple[NDArray, ...]:
        """
        Return the patch corresponding to the provided index.

        Parameters
        ----------
        index : int
            Index of the patch to return.

        Returns
        -------
        tuple(numpy.ndarray, ...)
            Transformed patch.
        """
        return self.patch_transform(patch=self.data[index])



from pathlib import Path

import numpy as np
import pytest
import zarr
from numpy.typing import NDArray

from careamics.dataset.dataset_utils import reshape_array
from careamics.dataset_ng.patch_extractor.image_stack import ZarrImageStack


def create_test_zarr(file_path: Path, data_path: str, data: NDArray):
    store = zarr.storage.FSStore(url=file_path.resolve())
    # create array
    array = zarr.create(
        store=store,
        shape=data.shape,
        chunks=data.shape,  # only 1 chunk
        dtype=np.uint16,
        path=data_path,
    )
    # write data
    array[...] = data
    store.close()


@pytest.mark.parametrize(
    "original_axes, original_shape, expected_shape, sample_idx",
    [
        ("YX", (32, 48), (1, 1, 32, 48), 0),
        ("XYS", (48, 32, 3), (3, 1, 32, 48), 1),
        ("SXYC", (3, 48, 32, 2), (3, 2, 32, 48), 1),
        ("CYXT", (2, 32, 48, 3), (3, 2, 32, 48), 2),
        ("CXYTS", (2, 48, 32, 3, 2), (6, 2, 32, 48), 4),
        ("XCSYT", (48, 1, 2, 32, 3), (6, 1, 32, 48), 5),  # crazy one
    ],
)
def test_extract_patch_2D(
    tmp_path: Path,
    original_axes: str,
    original_shape: tuple[int, ...],
    expected_shape: tuple[int, ...],
    sample_idx: int,
):
    # reference data to compare against, it is reshaped using careamics reshape_array
    data = np.arange(np.prod(original_shape)).reshape(original_shape)
    data_ref = reshape_array(data, original_axes)

    # save data as a zarr array to ininitialise image stack with
    file_path = tmp_path / "test_zarr.zarr"
    data_path = "image"
    create_test_zarr(file_path=file_path, data_path=data_path, data=data)

    # initialise ZarrImageStack
    store = zarr.storage.FSStore(url=file_path)
    image_stack = ZarrImageStack(store=store, data_path=data_path, axes=original_axes)
    # TODO: this assert can move if _reshaped_data_shape is tested separately
    assert image_stack.data_shape == expected_shape

    # test extracted patch matches patch from reference data
    coords = (11, 4)
    patch_size = (16, 9)

    extracted_patch = image_stack.extract_patch(
        sample_idx=sample_idx, coords=coords, patch_size=patch_size
    )
    patch_ref = data_ref[
        sample_idx,
        :,
        coords[0] : coords[0] + patch_size[0],
        coords[1] : coords[1] + patch_size[1],
    ]
    np.testing.assert_array_equal(extracted_patch, patch_ref)


def test_from_ome_zarr():
    # kinda an integration test
    path = "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0062A/6001240.zarr"
    image_stack = ZarrImageStack.from_ome_zarr(path=path)  # initialise image stack
    n_channels = image_stack.data_shape[1]
    patch_size = (100, 64, 25)
    patch = image_stack.extract_patch(
        sample_idx=0, coords=(112, 56, 15), patch_size=patch_size
    )
    assert isinstance(patch, np.ndarray)  # make sure patch is numpy
    assert patch.shape == (n_channels, *patch_size)  # extracted patch has expected size



"""In-memory tiled prediction dataset."""

from __future__ import annotations

from numpy.typing import NDArray
from torch.utils.data import Dataset

from careamics.transforms import Compose

from ..config import InferenceConfig
from ..config.tile_information import TileInformation
from ..config.transformations import NormalizeModel
from .dataset_utils import reshape_array
from .tiling import extract_tiles


class InMemoryTiledPredDataset(Dataset):
    """Prediction dataset storing data in memory and returning tiles of each image.

    Parameters
    ----------
    prediction_config : InferenceConfig
        Prediction configuration.
    inputs : NDArray
        Input data.
    """

    def __init__(
        self,
        prediction_config: InferenceConfig,
        inputs: NDArray,
    ) -> None:
        """Constructor.

        Parameters
        ----------
        prediction_config : InferenceConfig
            Prediction configuration.
        inputs : NDArray
            Input data.

        Raises
        ------
        ValueError
            If data_path is not a directory.
        """
        if (
            prediction_config.tile_size is None
            or prediction_config.tile_overlap is None
        ):
            raise ValueError(
                "Tile size and overlap must be provided to use the tiled prediction "
                "dataset."
            )

        self.pred_config = prediction_config
        self.input_array = inputs
        self.axes = self.pred_config.axes
        self.tile_size = prediction_config.tile_size
        self.tile_overlap = prediction_config.tile_overlap
        self.image_means = self.pred_config.image_means
        self.image_stds = self.pred_config.image_stds

        # Generate patches
        self.data = self._prepare_tiles()

        # get transforms
        self.patch_transform = Compose(
            transform_list=[
                NormalizeModel(image_means=self.image_means, image_stds=self.image_stds)
            ],
        )

    def _prepare_tiles(self) -> list[tuple[NDArray, TileInformation]]:
        """
        Iterate over data source and create an array of patches.

        Returns
        -------
        list of tuples of NDArray and TileInformation
            List of tiles and tile information.
        """
        # reshape array
        reshaped_sample = reshape_array(self.input_array, self.axes)

        # generate patches, which returns a generator
        patch_generator = extract_tiles(
            arr=reshaped_sample,
            tile_size=self.tile_size,
            overlaps=self.tile_overlap,
        )
        patches_list = list(patch_generator)

        if len(patches_list) == 0:
            raise ValueError("No tiles generated, ")

        return patches_list

    def __len__(self) -> int:
        """
        Return the length of the dataset.

        Returns
        -------
        int
            Length of the dataset.
        """
        return len(self.data)

    def __getitem__(self, index: int) -> tuple[tuple[NDArray, ...], TileInformation]:
        """
        Return the patch corresponding to the provided index.

        Parameters
        ----------
        index : int
            Index of the patch to return.

        Returns
        -------
        tuple of NDArray and TileInformation
            Transformed patch.
        """
        tile_array, tile_info = self.data[index]

        # Apply transforms
        transformed_tile = self.patch_transform(patch=tile_array)

        return transformed_tile, tile_info



import numpy as np
import pytest

from careamics.config.transformations import (
    NormalizeModel,
    XYFlipModel,
    XYRandomRotate90Model,
)
from careamics.transforms import Compose, XYFlip, XYRandomRotate90


def test_empty_compose(ordered_array):
    array = ordered_array((4, 5, 5))
    source = array[2:, ...]
    target = array[:2, ...]

    # instantiate Compose
    compose = Compose([])

    # apply the composed transform
    source_transformed, target_transformed = compose(source, target)

    # check the results
    assert (source_transformed == source).all()
    assert (target_transformed == target).all()


def test_compose_with_target(ordered_array):
    seed = 24
    array = ordered_array((4, 5, 5))
    source = array[2:, ...]
    target = array[:2, ...]

    # transform lists
    transform_list = [XYFlip(seed=seed), XYRandomRotate90(seed=seed)]
    transform_list_pydantic = [
        XYFlipModel(name="XYFlip", seed=seed),
        XYRandomRotate90Model(name="XYRandomRotate90", seed=seed),
    ]

    # instantiate Compose
    compose = Compose(transform_list_pydantic)

    # apply the composed transform
    source_transformed, target_transformed = compose(source, target)

    # apply them manually
    t1_source, t1_target, _ = transform_list[0](source, target)
    t2_source, t2_target, _ = transform_list[1](t1_source, t1_target)

    # check the results
    assert (source_transformed == t2_source).all()
    assert (target_transformed == t2_target).all()


@pytest.mark.parametrize(
    "shape",
    [
        # 2D
        (1, 2, 2),
        (2, 2, 2),
        # 3D
        (1, 2, 2, 2),
        (2, 2, 2, 2),
    ],
)
def test_random_composition(ordered_array, shape):
    """Test that all the transforms can be composed in arbitrary order.

    This exclude the N2VManipulate transform, which needs to be the last one.
    """
    rng = np.random.default_rng(seed=42)

    # create array
    array = ordered_array(shape)
    mask = ordered_array(shape) + 5

    # apply transform in random order
    for _ in range(10):
        flip_x = rng.choice([True, False])

        transforms = [
            NormalizeModel(
                image_means=[0.5 for _ in range(array.shape[0])],
                image_stds=[0.5 for _ in range(array.shape[0])],
                target_means=[0.5 for _ in range(array.shape[0])],
                target_stds=[0.5 for _ in range(array.shape[0])],
            ),
            XYFlipModel(flip_x=flip_x, seed=42),
            XYRandomRotate90Model(seed=42),
        ]

        # randomly sort the transforms
        rng.shuffle(transforms)

        # apply compose
        compose = Compose(transforms)
        result_array, result_mask = compose(array, mask)
        assert not np.array_equal(result_array, array)
        assert not np.array_equal(result_mask, mask)


def test_compose_additional_arrays(ordered_array):
    """Test additional arrays get tranformed in the same way as the patch."""
    seed = 24
    # create inputs
    shape = (2, 2, 5, 5)
    array = ordered_array(shape)
    additional_arrays = {"arr": ordered_array(shape)}

    # create tranforms
    transforms = [
        XYFlipModel(name="XYFlip", seed=seed),
        XYRandomRotate90Model(name="XYRandomRotate90", seed=seed),
    ]

    compose = Compose(transforms)

    augmented, _, additional_augmented = compose.transform_with_additional_arrays(
        array, **additional_arrays
    )
    assert np.array_equal(augmented, additional_augmented["arr"])



"""Sequential patching functions."""

from typing import Optional, Union

import numpy as np
from skimage.util import view_as_windows

from .validate_patch_dimension import validate_patch_dimensions


def _compute_number_of_patches(
    arr_shape: tuple[int, ...], patch_sizes: Union[list[int], tuple[int, ...]]
) -> tuple[int, ...]:
    """
    Compute the number of patches that fit in each dimension.

    Parameters
    ----------
    arr_shape : tuple[int, ...]
        Shape of the input array.
    patch_sizes : Union[list[int], tuple[int, ...]
        Shape of the patches.

    Returns
    -------
    tuple[int, ...]
        Number of patches in each dimension.
    """
    if len(arr_shape) != len(patch_sizes):
        raise ValueError(
            f"Array shape {arr_shape} and patch size {patch_sizes} should have the "
            f"same dimension, including singleton dimension for S and equal dimension "
            f"for C."
        )

    try:
        n_patches = [
            np.ceil(arr_shape[i] / patch_sizes[i]).astype(int)
            for i in range(len(patch_sizes))
        ]
    except IndexError as e:
        raise ValueError(
            f"Patch size {patch_sizes} is not compatible with array shape {arr_shape}"
        ) from e

    return tuple(n_patches)


def _compute_overlap(
    arr_shape: tuple[int, ...], patch_sizes: Union[list[int], tuple[int, ...]]
) -> tuple[int, ...]:
    """
    Compute the overlap between patches in each dimension.

    If the array dimensions are divisible by the patch sizes, then the overlap is
    0. Otherwise, it is the result of the division rounded to the upper value.

    Parameters
    ----------
    arr_shape : tuple[int, ...]
        Input array shape.
    patch_sizes : Union[list[int], tuple[int, ...]]
        Size of the patches.

    Returns
    -------
    tuple[int, ...]
        Overlap between patches in each dimension.
    """
    n_patches = _compute_number_of_patches(arr_shape, patch_sizes)

    overlap = [
        np.ceil(
            np.clip(n_patches[i] * patch_sizes[i] - arr_shape[i], 0, None)
            / max(1, (n_patches[i] - 1))
        ).astype(int)
        for i in range(len(patch_sizes))
    ]
    return tuple(overlap)


def _compute_patch_steps(
    patch_sizes: Union[list[int], tuple[int, ...]], overlaps: tuple[int, ...]
) -> tuple[int, ...]:
    """
    Compute steps between patches.

    Parameters
    ----------
    patch_sizes : tuple[int]
        Size of the patches.
    overlaps : tuple[int]
        Overlap between patches.

    Returns
    -------
    tuple[int]
        Steps between patches.
    """
    steps = [
        min(patch_sizes[i] - overlaps[i], patch_sizes[i])
        for i in range(len(patch_sizes))
    ]
    return tuple(steps)


# TODO why stack the target here and not on a different dimension before this function?
def _compute_patch_views(
    arr: np.ndarray,
    window_shape: list[int],
    step: tuple[int, ...],
    output_shape: list[int],
    target: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Compute views of an array corresponding to patches.

    Parameters
    ----------
    arr : np.ndarray
        Array from which the views are extracted.
    window_shape : tuple[int]
        Shape of the views.
    step : tuple[int]
        Steps between views.
    output_shape : tuple[int]
        Shape of the output array.
    target : Optional[np.ndarray], optional
        Target array, by default None.

    Returns
    -------
    np.ndarray
        Array with views dimension.
    """
    rng = np.random.default_rng()

    if target is not None:
        arr = np.stack([arr, target], axis=0)
        window_shape = [arr.shape[0], *window_shape]
        step = (arr.shape[0], *step)
        output_shape = [-1, arr.shape[0], arr.shape[2], *output_shape[2:]]

    patches = view_as_windows(arr, window_shape=window_shape, step=step).reshape(
        *output_shape
    )
    rng.shuffle(patches, axis=0)
    return patches


def extract_patches_sequential(
    arr: np.ndarray,
    patch_size: Union[list[int], tuple[int, ...]],
    target: Optional[np.ndarray] = None,
) -> tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Generate patches from an array in a sequential manner.

    Array dimensions should be SC(Z)YX, where S and C can be singleton dimensions. The
    patches are generated sequentially and cover the whole array.

    Parameters
    ----------
    arr : np.ndarray
        Input image array.
    patch_size : tuple[int]
        Patch sizes in each dimension.
    target : Optional[np.ndarray], optional
        Target array, by default None.

    Returns
    -------
    tuple[np.ndarray, Optional[np.ndarray]]
        Patches.
    """
    is_3d_patch = len(patch_size) == 3

    # Patches sanity check
    validate_patch_dimensions(arr, patch_size, is_3d_patch)

    # Update patch size to encompass S and C dimensions
    patch_size = [1, arr.shape[1], *patch_size]

    # Compute overlap
    overlaps = _compute_overlap(arr_shape=arr.shape, patch_sizes=patch_size)

    # Create view window and overlaps
    window_steps = _compute_patch_steps(patch_sizes=patch_size, overlaps=overlaps)

    output_shape = [
        -1,
    ] + patch_size[1:]

    # Generate a view of the input array containing pre-calculated number of patches
    # in each dimension with overlap.
    # Resulting array is resized to (n_patches, C, Z, Y, X) or (n_patches, C, Y, X)
    patches = _compute_patch_views(
        arr,
        window_shape=patch_size,
        step=window_steps,
        output_shape=output_shape,
        target=target,
    )

    if target is not None:
        # target was concatenated to patches in _compute_reshaped_view
        return (
            patches[:, 0, ...],
            patches[:, 1, ...],
        )  # TODO  in _compute_reshaped_view?
    else:
        return patches, None



"""Convenience function to create covers for the BMZ."""

from pathlib import Path

import numpy as np
from numpy.typing import NDArray
from PIL import Image

color_palette = np.array(
    [
        np.array([255, 195, 0]),  # grey
        np.array([189, 226, 240]),
        np.array([96, 60, 76]),
        np.array([193, 225, 193]),
    ]
)


def _get_norm_slice(array: NDArray) -> NDArray:
    """Get the normalized middle slice of a 4D or 5D array (SC(Z)YX).

    Parameters
    ----------
    array : NDArray
        Array from which to get the middle slice.

    Returns
    -------
    NDArray
        Normalized middle slice of the input array.
    """
    if array.ndim not in (4, 5):
        raise ValueError("Array must be 4D or 5D.")

    channels = array.shape[1] > 1
    z_stack = array.ndim == 5

    # get slice
    if z_stack:
        array_slice = array[0, :, array.shape[2] // 2, ...]
    else:
        array_slice = array[0, ...]

    # channels
    if channels:
        array_slice = np.moveaxis(array_slice, 0, -1)
    else:
        array_slice = array_slice[0, ...]

    # normalize
    array_slice = (
        255
        * (array_slice - array_slice.min())
        / (array_slice.max() - array_slice.min())
    )

    return array_slice.astype(np.uint8)


def _four_channel_image(array: NDArray) -> Image:
    """Convert 4-channel array to Image.

    Parameters
    ----------
    array : NDArray
        Normalized array to convert.

    Returns
    -------
    Image
        Converted array.
    """
    colors = color_palette[np.newaxis, np.newaxis, :, :]
    four_c_array = np.sum(array[..., :4, np.newaxis] * colors, axis=-2).astype(np.uint8)

    return Image.fromarray(four_c_array).convert("RGB")


def _convert_to_image(original_shape: tuple[int, ...], array: NDArray) -> Image:
    """Convert to Image.

    Parameters
    ----------
    original_shape : tuple
        Original shape of the array.
    array : NDArray
        Normalized array to convert.

    Returns
    -------
    Image
        Converted array.
    """
    n_channels = original_shape[1]

    if n_channels > 1:
        if n_channels == 3:
            return Image.fromarray(array).convert("RGB")
        elif n_channels == 2:
            # add an empty channel to the numpy array
            array = np.concatenate([np.zeros_like(array[..., 0:1]), array], axis=-1)

            return Image.fromarray(array).convert("RGB")
        else:  # more than 4
            return _four_channel_image(array[..., :4])
    else:
        return Image.fromarray(array).convert("L").convert("RGB")


def create_cover(directory: Path, array_in: NDArray, array_out: NDArray) -> Path:
    """Create a cover image from input and output arrays.

    Input and output arrays are expected to be SC(Z)YX. For images with a Z
    dimension, the middle slice is taken.

    Parameters
    ----------
    directory : Path
        Directory in which to save the cover.
    array_in : numpy.ndarray
        Array from which to create the cover image.
    array_out : numpy.ndarray
        Array from which to create the cover image.

    Returns
    -------
    Path
        Path to the saved cover image.
    """
    # extract slice and normalize arrays
    slice_in = _get_norm_slice(array_in)
    slice_out = _get_norm_slice(array_out)

    horizontal_split = slice_in.shape[-1] == slice_out.shape[-1]
    if not horizontal_split:
        if slice_in.shape[-2] != slice_out.shape[-2]:
            raise ValueError("Input and output arrays have different shapes.")

    # convert to Image
    image_in = _convert_to_image(array_in.shape, slice_in)
    image_out = _convert_to_image(array_out.shape, slice_out)

    # split horizontally or vertically
    if horizontal_split:
        width = image_in.width // 2

        cover = Image.new("RGB", (image_in.width, image_in.height))
        cover.paste(image_in.crop((0, 0, width, image_in.height)), (0, 0))
        cover.paste(
            image_out.crop(
                (image_in.width - width, 0, image_in.width, image_in.height)
            ),
            (width, 0),
        )
    else:
        height = image_in.height // 2

        cover = Image.new("RGB", (image_in.width, image_in.height))
        cover.paste(image_in.crop((0, 0, image_in.width, height)), (0, 0))
        cover.paste(
            image_out.crop(
                (0, image_in.height - height, image_in.width, image_in.height)
            ),
            (0, height),
        )

    # save
    cover_path = directory / "cover.png"
    cover.save(cover_path)

    return cover_path



import numpy as np
import pytest

from careamics.file_io.write import write_tiff


def test_write_tiff(tmp_path, ordered_array):
    """Test writing a tiff file."""
    # create an array
    array: np.ndarray = ordered_array((10, 10))

    # save files
    file = tmp_path / "test.tiff"
    write_tiff(file, array)

    assert file.is_file()


def test_invalid_extension_error(tmp_path, ordered_array):
    """Test error is raised when a path with an invalid extension is used."""
    # create an array
    array: np.ndarray = ordered_array((10, 10))

    # save files
    file = tmp_path / "test.invalid"
    with pytest.raises(ValueError):
        write_tiff(file, array)



"""Callback Pydantic models."""

from __future__ import annotations

from datetime import timedelta
from typing import Literal, Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
)


class CheckpointModel(BaseModel):
    """Checkpoint saving callback Pydantic model.

    The parameters corresponds to those of
    `pytorch_lightning.callbacks.ModelCheckpoint`.

    See:
    https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.callbacks.ModelCheckpoint.html#modelcheckpoint
    """

    model_config = ConfigDict(
        validate_assignment=True,
    )

    monitor: Literal["val_loss"] = Field(default="val_loss", validate_default=True)
    """Quantity to monitor."""

    verbose: bool = Field(default=False, validate_default=True)
    """Verbosity mode."""

    save_weights_only: bool = Field(default=False, validate_default=True)
    """When `True`, only the model's weights will be saved (model.save_weights)."""

    save_last: Optional[Literal[True, False, "link"]] = Field(
        default=True, validate_default=True
    )
    """When `True`, saves a last.ckpt copy whenever a checkpoint file gets saved."""

    save_top_k: int = Field(default=3, ge=1, le=10, validate_default=True)
    """If `save_top_k == kz, the best k models according to the quantity monitored
    will be saved. If `save_top_k == 0`, no models are saved. if `save_top_k == -1`,
    all models are saved."""

    mode: Literal["min", "max"] = Field(default="min", validate_default=True)
    """One of {min, max}. If `save_top_k != 0`, the decision to overwrite the current
    save file is made based on either the maximization or the minimization of the
    monitored quantity. For 'val_acc', this should be 'max', for 'val_loss' this should
    be 'min', etc.
    """

    auto_insert_metric_name: bool = Field(default=False, validate_default=True)
    """When `True`, the checkpoints filenames will contain the metric name."""

    every_n_train_steps: Optional[int] = Field(
        default=None, ge=1, le=10, validate_default=True
    )
    """Number of training steps between checkpoints."""

    train_time_interval: Optional[timedelta] = Field(
        default=None, validate_default=True
    )
    """Checkpoints are monitored at the specified time interval."""

    every_n_epochs: Optional[int] = Field(
        default=None, ge=1, le=10, validate_default=True
    )
    """Number of epochs between checkpoints."""


class EarlyStoppingModel(BaseModel):
    """Early stopping callback Pydantic model.

    The parameters corresponds to those of
    `pytorch_lightning.callbacks.ModelCheckpoint`.

    See:
    https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.callbacks.EarlyStopping.html#lightning.pytorch.callbacks.EarlyStopping
    """

    model_config = ConfigDict(
        validate_assignment=True,
    )

    monitor: Literal["val_loss"] = Field(default="val_loss", validate_default=True)
    """Quantity to monitor."""

    min_delta: float = Field(default=0.0, ge=0.0, le=1.0, validate_default=True)
    """Minimum change in the monitored quantity to qualify as an improvement, i.e. an
    absolute change of less than or equal to min_delta, will count as no improvement."""

    patience: int = Field(default=3, ge=1, le=10, validate_default=True)
    """Number of checks with no improvement after which training will be stopped."""

    verbose: bool = Field(default=False, validate_default=True)
    """Verbosity mode."""

    mode: Literal["min", "max", "auto"] = Field(default="min", validate_default=True)
    """One of {min, max, auto}."""

    check_finite: bool = Field(default=True, validate_default=True)
    """When `True`, stops training when the monitored quantity becomes `NaN` or
    `inf`."""

    stopping_threshold: Optional[float] = Field(default=None, validate_default=True)
    """Stop training immediately once the monitored quantity reaches this threshold."""

    divergence_threshold: Optional[float] = Field(default=None, validate_default=True)
    """Stop training as soon as the monitored quantity becomes worse than this
    threshold."""

    check_on_train_epoch_end: Optional[bool] = Field(
        default=False, validate_default=True
    )
    """Whether to run early stopping at the end of the training epoch. If this is
    `False`, then the check runs at the end of the validation."""

    log_rank_zero_only: bool = Field(default=False, validate_default=True)
    """When set `True`, logs the status of the early stopping callback only for rank 0
    process."""



"""Utility functions for paths."""

from pathlib import Path
from typing import Union


def check_path_exists(path: Union[str, Path]) -> Path:
    """Check if a path exists. If not, raise an error.

    Note that it returns `path` as a Path object.

    Parameters
    ----------
    path : Union[str, Path]
        Path to check.

    Returns
    -------
    Path
        Path as a Path object.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Data path {path} is incorrect or does not exist.")

    return path



"""Computing data statistics."""

import numpy as np
from numpy.typing import NDArray


def compute_normalization_stats(image: NDArray) -> tuple[NDArray, NDArray]:
    """
    Compute mean and standard deviation of an array.

    Expected input shape is (S, C, (Z), Y, X). The mean and standard deviation are
    computed per channel.

    Parameters
    ----------
    image : NDArray
        Input array.

    Returns
    -------
    tuple of (list of floats, list of floats)
        Lists of mean and standard deviation values per channel.
    """
    # Define the lists for storing mean and std values
    means, stds = [], []
    # Iterate over the channels dimension and compute mean and std
    for ax in range(image.shape[1]):
        means.append(image[:, ax, ...].mean())
        stds.append(image[:, ax, ...].std())
    return np.stack(means), np.stack(stds)


def update_iterative_stats(
    count: NDArray, mean: NDArray, m2: NDArray, new_values: NDArray
) -> tuple[NDArray, NDArray, NDArray]:
    """Update the mean and variance of an array iteratively.

    Parameters
    ----------
    count : NDArray
        Number of elements in the array. Shape: (C,).
    mean : NDArray
        Mean of the array. Shape: (C,).
    m2 : NDArray
        Variance of the array. Shape: (C,).
    new_values : NDArray
        New values to add to the mean and variance. Shape: (C, 1, 1, Z, Y, X).

    Returns
    -------
    tuple[NDArray, NDArray, NDArray]
        Updated count, mean, and variance.
    """
    num_channels = len(new_values)

    # --- update channel-wise counts ---
    count += np.ones_like(count) * np.prod(new_values.shape[1:])

    # --- update channel-wise mean ---
    # compute (new_values - old_mean) -> shape: (C, Z*Y*X)
    delta = new_values.reshape(num_channels, -1) - mean.reshape(num_channels, 1)
    mean += np.sum(delta / count.reshape(num_channels, 1), axis=1)

    # --- update channel-wise SoS ---
    # compute (new_values - new_mean) -> shape: (C, Z*Y*X)
    delta2 = new_values.reshape(num_channels, -1) - mean.reshape(num_channels, 1)
    m2 += np.sum(delta * delta2, axis=1)

    return count, mean, m2


def finalize_iterative_stats(
    count: NDArray, mean: NDArray, m2: NDArray
) -> tuple[NDArray, NDArray]:
    """Finalize the mean and variance computation.

    Parameters
    ----------
    count : NDArray
        Number of elements in the array. Shape: (C,).
    mean : NDArray
        Mean of the array. Shape: (C,).
    m2 : NDArray
        Variance of the array. Shape: (C,).

    Returns
    -------
    tuple[NDArray, NDArray]
        Final channel-wise mean and standard deviation.
    """
    std = np.sqrt(m2 / count)
    if any(c < 2 for c in count):
        return np.full(mean.shape, np.nan), np.full(std.shape, np.nan)
    else:
        return mean, std


class WelfordStatistics:
    """Compute Welford statistics iteratively.

    The Welford algorithm is used to compute the mean and variance of an array
    iteratively. Based on the implementation from:
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm
    """

    def update(self, array: NDArray, sample_idx: int) -> None:
        """Update the Welford statistics.

        Parameters
        ----------
        array : NDArray
            Input array.
        sample_idx : int
            Current sample number.
        """
        self.sample_idx = sample_idx
        sample_channels = np.array(np.split(array, array.shape[1], axis=1))

        # Initialize the statistics
        if self.sample_idx == 0:
            # Compute the mean and standard deviation
            self.mean, _ = compute_normalization_stats(array)
            # Initialize the count and m2 with zero-valued arrays of shape (C,)
            self.count, self.mean, self.m2 = update_iterative_stats(
                count=np.zeros(array.shape[1]),
                mean=self.mean,
                m2=np.zeros(array.shape[1]),
                new_values=sample_channels,
            )
        else:
            # Update the statistics
            self.count, self.mean, self.m2 = update_iterative_stats(
                count=self.count, mean=self.mean, m2=self.m2, new_values=sample_channels
            )

        self.sample_idx += 1

    def finalize(self) -> tuple[NDArray, NDArray]:
        """Finalize the Welford statistics.

        Returns
        -------
        tuple or numpy arrays
            Final mean and standard deviation.
        """
        return finalize_iterative_stats(self.count, self.mean, self.m2)


# from multiprocessing import Value
# from typing import tuple

# import numpy as np


# class RunningStats:
#     """Calculates running mean and std."""

#     def __init__(self) -> None:
#         self.reset()

#     def reset(self) -> None:
#         """Reset the running stats."""
#         self.avg_mean = Value("d", 0)
#         self.avg_std = Value("d", 0)
#         self.m2 = Value("d", 0)
#         self.count = Value("i", 0)

#     def init(self, mean: float, std: float) -> None:
#         """Initialize running stats."""
#         with self.avg_mean.get_lock():
#             self.avg_mean.value += mean
#         with self.avg_std.get_lock():
#             self.avg_std.value = std

#     def compute_std(self) -> tuple[float, float]:
#         """Compute std."""
#         if self.count.value >= 2:
#             self.avg_std.value = np.sqrt(self.m2.value / self.count.value)

#     def update(self, value: float) -> None:
#         """Update running stats."""
#         with self.count.get_lock():
#             self.count.value += 1
#         delta = value - self.avg_mean.value
#         with self.avg_mean.get_lock():
#             self.avg_mean.value += delta / self.count.value
#         delta2 = value - self.avg_mean.value
#         with self.m2.get_lock():
#             self.m2.value += delta * delta2



import pytest

from careamics.config.optimizer_models import LrSchedulerModel, OptimizerModel
from careamics.config.support.supported_optimizers import (
    SupportedOptimizer,
    SupportedScheduler,
)


@pytest.mark.parametrize(
    "optimizer_name, parameters",
    [
        (
            SupportedOptimizer.ADAM.value,
            {
                "lr": 0.08,
                "betas": (0.1, 0.11),
                "eps": 6e-08,
                "weight_decay": 0.2,
                "amsgrad": True,
            },
        ),
        (
            SupportedOptimizer.SGD.value,
            {
                "lr": 0.11,
                "momentum": 5,
                "dampening": 1,
                "weight_decay": 8,
                "nesterov": True,
            },
        ),
    ],
)
def test_optimizer_parameters(optimizer_name: SupportedOptimizer, parameters: dict):
    """Test optimizer parameters filtering.

    For parameters, see:
    https://pytorch.org/docs/stable/optim.html#algorithms
    """
    # add non valid parameter
    new_parameters = parameters.copy()
    new_parameters["some_random_one"] = 42

    # create optimizer and check that the parameters are filtered
    optimizer = OptimizerModel(name=optimizer_name, parameters=new_parameters)
    assert optimizer.parameters == parameters


def test_sgd_missing_parameter():
    """Test that SGD optimizer fails if `lr` is not provided.

    Note: The SGD optimizer requires the `lr` parameter.
    """
    with pytest.raises(ValueError):
        OptimizerModel(name=SupportedOptimizer.SGD.value, parameters={})

    # test that it works if lr is provided
    optimizer = OptimizerModel(
        name=SupportedOptimizer.SGD.value, parameters={"lr": 0.1}
    )
    assert optimizer.parameters == {"lr": 0.1}


def test_optimizer_wrong_values_by_assignments():
    """Test that wrong values cause an error during assignment."""
    optimizer = OptimizerModel(
        name=SupportedOptimizer.ADAM.value, parameters={"lr": 0.08}
    )

    # name
    optimizer.name = SupportedOptimizer.SGD.value
    with pytest.raises(ValueError):
        optimizer.name = "MyOptim"

    # parameters
    optimizer.parameters = {"lr": 0.1}
    with pytest.raises(ValueError):
        optimizer.parameters = "lr = 0.3"


def test_optimizer_to_dict_optional():
    """ "Test that export to dict includes optional values."""
    config = {
        "name": "Adam",
        "parameters": {
            "lr": 0.1,
            "betas": (0.1, 0.11),
        },
    }

    optim_minimum = OptimizerModel(**config).model_dump()
    assert optim_minimum == config


@pytest.mark.parametrize(
    "lr_scheduler_name, parameters",
    [
        (
            SupportedScheduler.REDUCE_LR_ON_PLATEAU.value,
            {
                "mode": "max",
                "factor": 0.3,
                "patience": 5,
                "threshold": 0.003,
                "threshold_mode": "abs",
                "cooldown": 3,
                "min_lr": 0.1,
                "eps": 5e-08,
            },
        ),
        (
            SupportedScheduler.STEP_LR.value,
            {
                "step_size": 2,
                "gamma": 0.3,
                "last_epoch": -5,
            },
        ),
    ],
)
def test_scheduler_parameters(lr_scheduler_name: SupportedScheduler, parameters: dict):
    """Test lr scheduler parameters filtering.

    For parameters, see:
    https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate
    """
    # add non valid parameter
    new_parameters = parameters.copy()
    new_parameters["some_random_one"] = 42

    # create optimizer and check that the parameters are filtered
    lr_scheduler = LrSchedulerModel(name=lr_scheduler_name, parameters=new_parameters)
    assert lr_scheduler.parameters == parameters


def test_scheduler_missing_parameter():
    """Test that StepLR scheduler fails if `step_size` is not provided"""
    with pytest.raises(ValueError):
        LrSchedulerModel(name=SupportedScheduler.STEP_LR.value, parameters={})

    # test that it works if lr is provided
    lr_scheduler = LrSchedulerModel(
        name=SupportedScheduler.STEP_LR.value, parameters={"step_size": "5"}
    )
    assert lr_scheduler.parameters == {"step_size": "5"}



"""Architectures supported by CAREamics."""

from careamics.utils import BaseEnum


class SupportedArchitecture(str, BaseEnum):
    """Supported architectures."""

    UNET = "UNet"
    """UNet architecture used with N2V, CARE and Noise2Noise."""

    LVAE = "LVAE"
    """Ladder Variational Autoencoder used for muSplit and denoiSplit."""



"""
Script for utility functions needed by the LVAE model.
"""

from typing import Literal, Sequence

import numpy as np
import torch
import torch.nn as nn
import torchvision.transforms.functional as F
from torch.distributions.normal import Normal


def torch_nanmean(inp):
    return torch.mean(inp[~inp.isnan()])


def power_of_2(self, x):
    assert isinstance(x, int)
    if x == 1:
        return True
    if x == 0:
        # happens with validation
        return False
    if x % 2 == 1:
        return False
    return self.power_of_2(x // 2)


class Enum:
    @classmethod
    def name(cls, enum_type):
        for key, value in cls.__dict__.items():
            if enum_type == value:
                return key

    @classmethod
    def contains(cls, enum_type):
        for key, value in cls.__dict__.items():
            if enum_type == value:
                return True
        return False

    @classmethod
    def from_name(cls, enum_type_str):
        for key, value in cls.__dict__.items():
            if key == enum_type_str:
                return value
        assert f"{cls.__name__}:{enum_type_str} doesnot exist."


class LossType(Enum):
    Elbo = 0
    ElboWithCritic = 1
    ElboMixedReconstruction = 2
    MSE = 3
    ElboWithNbrConsistency = 4
    ElboSemiSupMixedReconstruction = 5
    ElboCL = 6
    ElboRestrictedReconstruction = 7
    DenoiSplitMuSplit = 8


class ModelType(Enum):
    LadderVae = 3
    LadderVaeTwinDecoder = 4
    LadderVAECritic = 5
    # Separate vampprior: two optimizers
    LadderVaeSepVampprior = 6
    # one encoder for mixed input, two for separate inputs.
    LadderVaeSepEncoder = 7
    LadderVAEMultiTarget = 8
    LadderVaeSepEncoderSingleOptim = 9
    UNet = 10
    BraveNet = 11
    LadderVaeStitch = 12
    LadderVaeSemiSupervised = 13
    LadderVaeStitch2Stage = 14  # Note that previously trained models will have issue.
    # since earlier, LadderVaeStitch2Stage = 13, LadderVaeSemiSupervised = 14
    LadderVaeMixedRecons = 15
    LadderVaeCL = 16
    LadderVaeTwoDataSet = (
        17  # on one subdset, apply disentanglement, on other apply reconstruction
    )
    LadderVaeTwoDatasetMultiBranch = 18
    LadderVaeTwoDatasetMultiOptim = 19
    LVaeDeepEncoderIntensityAug = 20
    AutoRegresiveLadderVAE = 21
    LadderVAEInterleavedOptimization = 22
    Denoiser = 23
    DenoiserSplitter = 24
    SplitterDenoiser = 25
    LadderVAERestrictedReconstruction = 26
    LadderVAETwoDataSetRestRecon = 27
    LadderVAETwoDataSetFinetuning = 28


def _pad_crop_img(
    x: torch.Tensor, size: Sequence[int], mode: Literal["crop", "pad"]
) -> torch.Tensor:
    """Pads or crops a tensor.

    Pads or crops a tensor of shape (B, C, [Z], Y, X) to new shape.

    Parameters:
    -----------
    x: torch.Tensor
        Input image of shape (B, C, [Z], Y, X)
    size: Sequence[int]
        Desired size ([Z*], Y*, X*)
    mode: Literal["crop", "pad"]
        Mode, either 'pad' or 'crop'

    Returns:
    --------
    torch.Tensor:
        The padded or cropped tensor
    """
    # TODO: Support cropping/padding on selected dimensions
    assert (x.dim() == 4 and len(size) == 2) or (x.dim() == 5 and len(size) == 3)

    size = tuple(size)
    x_size = x.size()[2:]

    if mode == "pad":
        cond = any(x_size[i] > size[i] for i in range(len(size)))
    elif mode == "crop":
        cond = any(x_size[i] < size[i] for i in range(len(size)))

    if cond:
        raise ValueError(f"Trying to {mode} from size {x_size} to size {size}")

    diffs = [abs(x - s) for x, s in zip(x_size, size)]
    d1 = [d // 2 for d in diffs]
    d2 = [d - (d // 2) for d in diffs]

    if mode == "pad":
        if x.dim() == 4:
            padding = [d1[1], d2[1], d1[0], d2[0], 0, 0, 0, 0]
        elif x.dim() == 5:
            padding = [d1[2], d2[2], d1[1], d2[1], d1[0], d2[0], 0, 0, 0, 0]
        return nn.functional.pad(x, padding)
    elif mode == "crop":
        if x.dim() == 4:
            return x[:, :, d1[0] : (x_size[0] - d2[0]), d1[1] : (x_size[1] - d2[1])]
        elif x.dim() == 5:
            return x[
                :,
                :,
                d1[0] : (x_size[0] - d2[0]),
                d1[1] : (x_size[1] - d2[1]),
                d1[2] : (x_size[2] - d2[2]),
            ]


def pad_img_tensor(x: torch.Tensor, size: Sequence[int]) -> torch.Tensor:
    """Pads a tensor

    Pads a tensor of shape (B, C, [Z], Y, X) to desired spatial dimensions.

    Parameters:
    -----------
        x (torch.Tensor): Input image of shape (B, C, [Z], Y, X)
        size (list or tuple): Desired size  ([Z*], Y*, X*)

    Returns:
    --------
        The padded tensor
    """
    return _pad_crop_img(x, size, "pad")


def crop_img_tensor(x, size) -> torch.Tensor:
    """Crops a tensor.
    Crops a tensor of shape (batch, channels, h, w) to a desired height and width
    given by a tuple.
    Args:
        x (torch.Tensor): Input image
        size (list or tuple): Desired size (height, width)

    Returns
    -------
        The cropped tensor
    """
    return _pad_crop_img(x, size, "crop")


class StableExponential:
    """
    Class that redefines the definition of exp() to increase numerical stability.
    Naturally, also the definition of log() must change accordingly.
    However, it is worth noting that the two operations remain one the inverse of the other,
    meaning that x = log(exp(x)) and x = exp(log(x)) are always true.

    Definition:
        exp(x) = {
            exp(x) if x<=0
            x+1    if x>0
        }

        log(x) = {
            x        if x<=0
            log(1+x) if x>0
        }

    NOTE 1:
        Within the class everything is done on the tensor given as input to the constructor.
        Therefore, when exp() is called, self._tensor.exp() is computed.
        When log() is called, torch.log(self._tensor.exp()) is computed instead.

    NOTE 2:
        Given the output from exp(), torch.log() or the log() method of the class give identical results.
    """

    def __init__(self, tensor):
        self._raw_tensor = tensor
        posneg_dic = self.posneg_separation(self._raw_tensor)
        self.pos_f, self.neg_f = posneg_dic["filter"]
        self.pos_data, self.neg_data = posneg_dic["value"]

    def posneg_separation(self, tensor):
        pos = tensor > 0
        pos_tensor = torch.clip(tensor, min=0)

        neg = tensor <= 0
        neg_tensor = torch.clip(tensor, max=0)

        return {"filter": [pos, neg], "value": [pos_tensor, neg_tensor]}

    def exp(self):
        return torch.exp(self.neg_data) * self.neg_f + (1 + self.pos_data) * self.pos_f

    def log(self):
        return self.neg_data * self.neg_f + torch.log(1 + self.pos_data) * self.pos_f


class StableLogVar:
    """
    Class that provides a numerically stable implementation of Log-Variance.
    Specifically, it uses the exp() and log() formulas defined in `StableExponential` class.
    """

    def __init__(
        self, logvar: torch.Tensor, enable_stable: bool = True, var_eps: float = 1e-6
    ):
        """
        Constructor.

        Parameters
        ----------
        logvar: torch.Tensor
            The input (true) logvar vector, to be converted in the Stable version.
        enable_stable: bool, optional
            Whether to compute the stable version of log-variance. Default is `True`.
        var_eps: float, optional
            The minimum value attainable by the variance. Default is `1e-6`.
        """
        self._lv = logvar
        self._enable_stable = enable_stable
        self._eps = var_eps

    def get(self) -> torch.Tensor:
        if self._enable_stable is False:
            return self._lv

        return torch.log(self.get_var())

    def get_var(self) -> torch.Tensor:
        """
        Get Variance from Log-Variance.
        """
        if self._enable_stable is False:
            return torch.exp(self._lv)
        return StableExponential(self._lv).exp() + self._eps

    def get_std(self) -> torch.Tensor:
        return torch.sqrt(self.get_var())

    @property
    def is_3D(self) -> bool:
        """Check if the _lv tensor is 3D.

        Recall that, in this framework, tensors have shape (B, C, [Z], Y, X).
        """
        return self._lv.dim() == 5

    def centercrop_to_size(self, size: Sequence[int]) -> None:
        """
        Centercrop the log-variance tensor to the desired size.

        Parameters
        ----------
        size: torch.Tensor
            The desired size of the log-variance tensor.
        """
        assert not self.is_3D, "Centercrop is implemented only for 2D tensors."

        if self._lv.shape[-1] == size:
            return

        diff = self._lv.shape[-1] - size
        assert diff > 0 and diff % 2 == 0
        self._lv = F.center_crop(self._lv, (size, size))


class StableMean:

    def __init__(self, mean):
        self._mean = mean

    def get(self) -> torch.Tensor:
        return self._mean

    @property
    def is_3D(self) -> bool:
        """Check if the _mean tensor is 3D.

        Recall that, in this framework, tensors have shape (B, C, [Z], Y, X).
        """
        return self._mean.dim() == 5

    def centercrop_to_size(self, size: Sequence[int]) -> None:
        """Centercrop the mean tensor to the desired size.

        Implemented only in the case of 2D tensors.

        Parameters
        ----------
        size: torch.Tensor
            The desired size of the log-variance tensor.
        """
        assert not self.is_3D, "Centercrop is implemented only for 2D tensors."

        if self._mean.shape[-1] == size:
            return

        diff = self._mean.shape[-1] - size
        assert diff > 0 and diff % 2 == 0
        self._mean = F.center_crop(self._mean, (size, size))


def allow_numpy(func):
    """
    All optional arguments are passed as is. positional arguments are checked. if they are numpy array,
    they are converted to torch Tensor.
    """

    def numpy_wrapper(*args, **kwargs):
        new_args = []
        for arg in args:
            if isinstance(arg, np.ndarray):
                arg = torch.Tensor(arg)
            new_args.append(arg)
        new_args = tuple(new_args)

        output = func(*new_args, **kwargs)
        return output

    return numpy_wrapper


class Interpolate(nn.Module):
    """Wrapper for torch.nn.functional.interpolate."""

    def __init__(self, size=None, scale=None, mode="bilinear", align_corners=False):
        super().__init__()
        assert (size is None) == (scale is not None)
        self.size = size
        self.scale = scale
        self.mode = mode
        self.align_corners = align_corners

    def forward(self, x):
        out = F.interpolate(
            x,
            size=self.size,
            scale_factor=self.scale,
            mode=self.mode,
            align_corners=self.align_corners,
        )
        return out


def kl_normal_mc(z, p_mulv, q_mulv):
    """
    One-sample estimation of element-wise KL between two diagonal
    multivariate normal distributions. Any number of dimensions,
    broadcasting supported (be careful).
    :param z:
    :param p_mulv:
    :param q_mulv:
    :return:
    """
    assert isinstance(p_mulv, tuple)
    assert isinstance(q_mulv, tuple)
    p_mu, p_lv = p_mulv
    q_mu, q_lv = q_mulv

    p_std = p_lv.get_std()
    q_std = q_lv.get_std()

    p_distrib = Normal(p_mu.get(), p_std)
    q_distrib = Normal(q_mu.get(), q_std)
    return q_distrib.log_prob(z) - p_distrib.log_prob(z)



"""Functions to reimplement the tiling in the Disentangle repository."""

import builtins
import itertools
from collections.abc import Generator
from typing import Any, Optional, Union

import numpy as np
from numpy.typing import NDArray

from careamics.config.tile_information import TileInformation
from careamics.lvae_training.dataset.utils.index_manager import GridIndexManager


def extract_tiles(
    arr: NDArray,
    tile_size: NDArray[np.int_],
    overlaps: NDArray[np.int_],
    padding_kwargs: Optional[dict[str, Any]] = None,
) -> Generator[tuple[NDArray, TileInformation], None, None]:
    """Generate tiles from the input array with specified overlap.

    The tiles cover the whole array; which will be additionally padded, to ensure that
    the section of the tile that contributes to the final image comes from the center
    of the tile.

    The method returns a generator that yields tuples of array and tile information,
    the latter includes whether the tile is the last one, the coordinates of the
    overlap crop, and the coordinates of the stitched tile.

    Input array should have shape SC(Z)YX, while the returned tiles have shape C(Z)YX,
    where C can be a singleton.

    Parameters
    ----------
    arr : np.ndarray
        Array of shape (S, C, (Z), Y, X).
    tile_size : 1D numpy.ndarray of tuple
        Tile sizes in each dimension, of length 2 or 3.
    overlaps : 1D numpy.ndarray of tuple
        Overlap values in each dimension, of length 2 or 3.
    padding_kwargs : dict, optional
        The arguments of `np.pad` after the first two arguments, `array` and
        `pad_width`. If not specified the default will be `{"mode": "reflect"}`. See
        `numpy.pad` docs:
        https://numpy.org/doc/stable/reference/generated/numpy.pad.html.

    Yields
    ------
    Generator[Tuple[np.ndarray, TileInformation], None, None]
        Tile generator, yields the tile and additional information.
    """
    if padding_kwargs is None:
        padding_kwargs = {"mode": "reflect"}

    # Iterate over num samples (S)
    for sample_idx in range(arr.shape[0]):
        sample = arr[sample_idx, ...]
        data_shape = np.array(sample.shape)

        # add padding to ensure evenly spaced & overlapping tiles.
        spatial_padding = compute_padding(data_shape, tile_size, overlaps)
        padding = ((0, 0), *spatial_padding)
        sample = np.pad(sample, padding, **padding_kwargs)

        # The number of tiles in each dimension, should be of length 2 or 3
        tile_grid_shape = compute_tile_grid_shape(data_shape, tile_size, overlaps)
        # itertools.product is equivalent of nested loops

        stitch_size = tile_size - overlaps
        for tile_grid_indices in itertools.product(
            *[range(n) for n in tile_grid_shape]
        ):

            # calculate crop coordinates
            crop_coords_start = np.array(tile_grid_indices) * stitch_size
            crop_slices: tuple[Union[builtins.ellipsis, slice], ...] = (
                ...,
                *[
                    slice(coords, coords + extent)
                    for coords, extent in zip(crop_coords_start, tile_size)
                ],
            )
            tile = sample[crop_slices]

            tile_info = compute_tile_info(
                np.array(tile_grid_indices),
                np.array(data_shape),
                np.array(tile_size),
                np.array(overlaps),
                sample_idx,
            )
            # TODO: kinda weird this is a generator,
            #   -> doesn't really save memory ? Don't think there are any places the
            #    tiles are not exracted all at the same time.
            #   Although I guess it would make sense for a zarr tile extractor.
            yield tile, tile_info


def compute_tile_info_legacy(
    grid_index_manager: GridIndexManager, index: int
) -> TileInformation:
    """
    Compute the tile information for a tile at a given dataset index.

    Parameters
    ----------
    grid_index_manager : GridIndexManager
        The grid index manager that keeps track of tile locations.
    index : int
        The dataset index.

    Returns
    -------
    TileInformation
        Information that describes how to crop and stitch a tile to create a full image.

    Raises
    ------
    ValueError
        If `grid_index_manager.data_shape` does not have 4 or 5 dimensions.
    """
    data_shape = np.array(grid_index_manager.data_shape)
    if len(data_shape) == 5:
        n_spatial_dims = 3
    elif len(data_shape) == 4:
        n_spatial_dims = 2
    else:
        raise ValueError("Data shape must have 4 or 5 dimensions, equating to SC(Z)YX.")

    stitch_coords_start = np.array(
        grid_index_manager.get_location_from_dataset_idx(index)
    )
    stitch_coords_end = stitch_coords_start + np.array(grid_index_manager.grid_shape)

    tile_coords_start = stitch_coords_start - grid_index_manager.patch_offset()

    # --- replace out of bounds indices
    out_of_lower_bound = stitch_coords_start < 0
    out_of_upper_bound = stitch_coords_end > data_shape
    stitch_coords_start[out_of_lower_bound] = 0
    stitch_coords_end[out_of_upper_bound] = data_shape[out_of_upper_bound]

    # TODO: TilingMode not in current version
    # if grid_index_manager.tiling_mode == TilingMode.ShiftBoundary:
    #     for dim in range(len(stitch_coords_start)):
    #         if tile_coords_start[dim] == 0:
    #             stitch_coords_start[dim] = 0
    #         if tile_coords_end[dim] == grid_index_manager.data_shape[dim]:
    #             tile_coords_end [dim]= grid_index_manager.data_shape[dim]

    # --- calculate overlap crop coords
    overlap_crop_coords_start = stitch_coords_start - tile_coords_start
    overlap_crop_coords_end = overlap_crop_coords_start + (
        stitch_coords_end - stitch_coords_start
    )

    last_tile = index == grid_index_manager.total_grid_count() - 1

    # --- combine start and end
    stitch_coords = tuple(
        (start, end) for start, end in zip(stitch_coords_start, stitch_coords_end)
    )
    overlap_crop_coords = tuple(
        (start, end)
        for start, end in zip(overlap_crop_coords_start, overlap_crop_coords_end)
    )

    tile_info = TileInformation(
        array_shape=data_shape[1:],  # remove S dim
        last_tile=last_tile,
        overlap_crop_coords=overlap_crop_coords[-n_spatial_dims:],
        stitch_coords=stitch_coords[-n_spatial_dims:],
        sample_id=0,
    )
    return tile_info


def compute_tile_info(
    tile_grid_indices: NDArray[np.int_],
    data_shape: NDArray[np.int_],
    tile_size: NDArray[np.int_],
    overlaps: NDArray[np.int_],
    sample_id: int = 0,
) -> TileInformation:
    """
    Compute the tile information for a tile with the coordinates `tile_grid_indices`.

    Parameters
    ----------
    tile_grid_indices : 1D np.array of int
        The coordinates of the tile within the tile grid, ((Z), Y, X), i.e. for 2D
        tiling the coordinates for the second tile in the first row of tiles would be
        (0, 1).
    data_shape : 1D np.array of int
        The shape of the data, should be (C, (Z), Y, X) where Z is optional.
    tile_size : 1D np.array of int
        Tile sizes in each dimension, of length 2 or 3.
    overlaps : 1D np.array of int
        Overlap values in each dimension, of length 2 or 3.
    sample_id : int, default=0
        An ID to identify which sample a tile belongs to.

    Returns
    -------
    TileInformation
        Information that describes how to crop and stitch a tile to create a full image.
    """
    spatial_dims_shape = data_shape[-len(tile_size) :]

    # The extent of the tile which will make up part of the stitched image.
    stitch_size = tile_size - overlaps
    stitch_coords_start = tile_grid_indices * stitch_size
    stitch_coords_end = stitch_coords_start + stitch_size

    tile_coords_start = stitch_coords_start - overlaps // 2

    # --- replace out of bounds indices
    out_of_lower_bound = stitch_coords_start < 0
    out_of_upper_bound = stitch_coords_end > spatial_dims_shape
    stitch_coords_start[out_of_lower_bound] = 0
    stitch_coords_end[out_of_upper_bound] = spatial_dims_shape[out_of_upper_bound]

    # --- calculate overlap crop coords
    overlap_crop_coords_start = stitch_coords_start - tile_coords_start
    overlap_crop_coords_end = overlap_crop_coords_start + (
        stitch_coords_end - stitch_coords_start
    )

    # --- combine start and end
    stitch_coords = tuple(
        (start, end) for start, end in zip(stitch_coords_start, stitch_coords_end)
    )
    overlap_crop_coords = tuple(
        (start, end)
        for start, end in zip(overlap_crop_coords_start, overlap_crop_coords_end)
    )

    # --- Check if last tile
    tile_grid_shape = np.array(compute_tile_grid_shape(data_shape, tile_size, overlaps))
    last_tile = (tile_grid_indices == (tile_grid_shape - 1)).all()

    tile_info = TileInformation(
        array_shape=data_shape,
        last_tile=last_tile,
        overlap_crop_coords=overlap_crop_coords,
        stitch_coords=stitch_coords,
        sample_id=sample_id,
    )
    return tile_info


def compute_padding(
    data_shape: NDArray[np.int_],
    tile_size: NDArray[np.int_],
    overlaps: NDArray[np.int_],
) -> tuple[tuple[int, int], ...]:
    """
    Calculate padding to ensure stitched data comes from the center of a tile.

    Padding is added to an array with shape `data_shape` so that when tiles are
    stitched together, the data used always comes from the center of a tile, even for
    tiles at the boundaries of the array.

    Parameters
    ----------
    data_shape : 1D numpy.array of int
        The shape of the data to be tiled and stitched together, (S, C, (Z), Y, X).
    tile_size : 1D numpy.array of int
        The tile size in each dimension, ((Z), Y, X).
    overlaps : 1D numpy.array of int
        The tile overlap in each dimension, ((Z), Y, X).

    Returns
    -------
    tuple of (int, int)
        A tuple specifying the padding to add in each dimension, each element is a two
        element tuple specifying the padding to add before and after the data. This
        can be used as the `pad_width` argument to `numpy.pad`.
    """
    tile_grid_shape = np.array(compute_tile_grid_shape(data_shape, tile_size, overlaps))
    covered_shape = (tile_size - overlaps) * tile_grid_shape + overlaps

    pad_before = overlaps // 2
    pad_after = covered_shape - data_shape[-len(tile_size) :] - pad_before

    return tuple((before, after) for before, after in zip(pad_before, pad_after))


def n_tiles_1d(axis_size: int, tile_size: int, overlap: int) -> int:
    """Calculate the number of tiles in a specific dimension.

    Parameters
    ----------
    axis_size : int
        The length of the data for in a specific dimension.
    tile_size : int
        The length of the tiles in a specific dimension.
    overlap : int
        The tile overlap in a specific dimension.

    Returns
    -------
    int
        The number of tiles that fit in one dimension given the arguments.
    """
    return int(np.ceil(axis_size / (tile_size - overlap)))


def total_n_tiles(
    data_shape: tuple[int, ...], tile_size: tuple[int, ...], overlaps: tuple[int, ...]
) -> int:
    """Calculate The total number of tiles over all dimensions.

    Parameters
    ----------
    data_shape : 1D numpy.array of int
        The shape of the data to be tiled and stitched together, (S, C, (Z), Y, X).
    tile_size : 1D numpy.array of int
        The tile size in each dimension, ((Z), Y, X).
    overlaps : 1D numpy.array of int
        The tile overlap in each dimension, ((Z), Y, X).


    Returns
    -------
    int
        The total number of tiles over all dimensions.
    """
    result = 1
    # assume spatial dimension are the last dimensions so iterate backwards
    for i in range(-1, -len(tile_size) - 1, -1):
        result = result * n_tiles_1d(data_shape[i], tile_size[i], overlaps[i])

    return result


def compute_tile_grid_shape(
    data_shape: NDArray[np.int_],
    tile_size: NDArray[np.int_],
    overlaps: NDArray[np.int_],
) -> tuple[int, ...]:
    """Calculate the number of tiles in each dimension.

    This can be thought of as a grid of tiles.

    Parameters
    ----------
    data_shape : 1D numpy.array of int
        The shape of the data to be tiled and stitched together, (S, C, (Z), Y, X).
    tile_size : 1D numpy.array of int
        The tile size in each dimension, ((Z), Y, X).
    overlaps : 1D numpy.array of int
        The tile overlap in each dimension, ((Z), Y, X).

    Returns
    -------
    tuple of int
        The number of tiles in each direction, ((Z, Y, X)).
    """
    shape = [0 for _ in range(len(tile_size))]
    # assume spatial dimension are the last dimensions so iterate backwards
    for i in range(-1, -len(tile_size) - 1, -1):
        shape[i] = n_tiles_1d(data_shape[i], tile_size[i], overlaps[i])
    return tuple(shape)



__all__ = [
    "ImageStack",
    "InMemoryImageStack",
    "ZarrImageStack",
]

from .image_stack_protocol import ImageStack
from .in_memory_image_stack import InMemoryImageStack
from .zarr_image_stack import ZarrImageStack



"""File utilities."""

from fnmatch import fnmatch
from pathlib import Path
from typing import Union

import numpy as np

from careamics.config.support import SupportedData
from careamics.utils.logging import get_logger

logger = get_logger(__name__)


def get_files_size(files: list[Path]) -> float:
    """Get files size in MB.

    Parameters
    ----------
    files : list of pathlib.Path
        List of files.

    Returns
    -------
    float
        Total size of the files in MB.
    """
    return np.sum([f.stat().st_size / 1024**2 for f in files])


def list_files(
    data_path: Union[str, Path],
    data_type: Union[str, SupportedData],
    extension_filter: str = "",
) -> list[Path]:
    """List recursively files in `data_path` and return a sorted list.

    If `data_path` is a file, its name is validated against the `data_type` using
    `fnmatch`, and the method returns `data_path` itself.

    By default, if `data_type` is equal to `custom`, all files will be listed. To
    further filter the files, use `extension_filter`.

    `extension_filter` must be compatible with `fnmatch` and `Path.rglob`, e.g. "*.npy"
    or "*.czi".

    Parameters
    ----------
    data_path : Union[str, Path]
        Path to the folder containing the data.
    data_type : Union[str, SupportedData]
        One of the supported data type (e.g. tif, custom).
    extension_filter : str, optional
        Extension filter, by default "".

    Returns
    -------
    list[Path]
        list of pathlib.Path objects.

    Raises
    ------
    FileNotFoundError
        If the data path does not exist.
    ValueError
        If the data path is empty or no files with the extension were found.
    ValueError
        If the file does not match the requested extension.
    """
    # convert to Path
    data_path = Path(data_path)

    # raise error if does not exists
    if not data_path.exists():
        raise FileNotFoundError(f"Data path {data_path} does not exist.")

    # get extension compatible with fnmatch and rglob search
    extension = SupportedData.get_extension_pattern(data_type)

    if data_type == SupportedData.CUSTOM and extension_filter != "":
        extension = extension_filter

    # search recurively
    if data_path.is_dir():
        # search recursively the path for files with the extension
        files = sorted(data_path.rglob(extension))
    else:
        # raise error if it has the wrong extension
        if not fnmatch(str(data_path.absolute()), extension):
            raise ValueError(
                f"File {data_path} does not match the requested extension "
                f'"{extension}".'
            )

        # save in list
        files = [data_path]

    # raise error if no files were found
    if len(files) == 0:
        raise ValueError(
            f'Data path {data_path} is empty or files with extension "{extension}" '
            f"were not found."
        )

    return files


def validate_source_target_files(src_files: list[Path], tar_files: list[Path]) -> None:
    """
    Validate source and target path lists.

    The two lists should have the same number of files, and the filenames should match.

    Parameters
    ----------
    src_files : list of pathlib.Path
        List of source files.
    tar_files : list of pathlib.Path
        List of target files.

    Raises
    ------
    ValueError
        If the number of files in source and target folders is not the same.
    ValueError
        If some filenames in Train and target folders are not the same.
    """
    # check equal length
    if len(src_files) != len(tar_files):
        raise ValueError(
            f"The number of source files ({len(src_files)}) is not equal to the number "
            f"of target files ({len(tar_files)})."
        )

    # check identical names
    src_names = {f.name for f in src_files}
    tar_names = {f.name for f in tar_files}
    difference = src_names.symmetric_difference(tar_names)

    if len(difference) > 0:
        raise ValueError(f"Source and target files have different names: {difference}.")



"""Functions to read tiff images."""

import logging
from fnmatch import fnmatch
from pathlib import Path

import numpy as np
import tifffile

from careamics.config.support import SupportedData
from careamics.utils.logging import get_logger

logger = get_logger(__name__)


def read_tiff(file_path: Path, *args: list, **kwargs: dict) -> np.ndarray:
    """
    Read a tiff file and return a numpy array.

    Parameters
    ----------
    file_path : Path
        Path to a file.
    *args : list
        Additional arguments.
    **kwargs : dict
        Additional keyword arguments.

    Returns
    -------
    np.ndarray
        Resulting array.

    Raises
    ------
    ValueError
        If the file failed to open.
    OSError
        If the file failed to open.
    ValueError
        If the file is not a valid tiff.
    ValueError
        If the data dimensions are incorrect.
    ValueError
        If the axes length is incorrect.
    """
    if fnmatch(
        file_path.suffix, SupportedData.get_extension_pattern(SupportedData.TIFF)
    ):
        try:
            array = tifffile.imread(file_path)
        except (ValueError, OSError) as e:
            logging.exception(f"Exception in file {file_path}: {e}, skipping it.")
            raise e
    else:
        raise ValueError(f"File {file_path} is not a valid tiff.")

    return array



"""Likelihood model."""

from typing import Annotated, Literal, Optional, Union

import numpy as np
import torch
from pydantic import BaseModel, ConfigDict, PlainSerializer, PlainValidator

from careamics.models.lvae.noise_models import (
    GaussianMixtureNoiseModel,
    MultiChannelNoiseModel,
)
from careamics.utils.serializers import _array_to_json, _to_torch

NoiseModel = Union[GaussianMixtureNoiseModel, MultiChannelNoiseModel]

# TODO: this is a temporary solution to serialize and deserialize tensor fields
# in pydantic models. Specifically, the aim is to enable saving and loading configs
# with such tensors to/from JSON files during, resp., training and evaluation.
Tensor = Annotated[
    Union[np.ndarray, torch.Tensor],
    PlainSerializer(_array_to_json, return_type=str),
    PlainValidator(_to_torch),
]
"""Annotated tensor type, used to serialize arrays or tensors to JSON strings
and deserialize them back to tensors."""


class GaussianLikelihoodConfig(BaseModel):
    """Gaussian likelihood configuration."""

    model_config = ConfigDict(validate_assignment=True)

    predict_logvar: Optional[Literal["pixelwise"]] = None
    """If `pixelwise`, log-variance is computed for each pixel, else log-variance
    is not computed."""

    logvar_lowerbound: Union[float, None] = None
    """The lowerbound value for log-variance."""


class NMLikelihoodConfig(BaseModel):
    """Noise model likelihood configuration.

    NOTE: we need to define the data mean and std here because the noise model
    is trained on not-normalized data. Hence, we need to unnormalize the model
    output to compute the noise model likelihood.
    """

    model_config = ConfigDict(validate_assignment=True, arbitrary_types_allowed=True)

    # TODO remove and use as parameters to the likelihood functions?
    data_mean: Tensor = torch.zeros(1)
    """The mean of the data, used to unnormalize data for noise model evaluation.
    Shape is (target_ch,) (or (1, target_ch, [1], 1, 1))."""

    # TODO remove and use as parameters to the likelihood functions?
    data_std: Tensor = torch.ones(1)
    """The standard deviation of the data, used to unnormalize data for noise
    model evaluation. Shape is (target_ch,) (or (1, target_ch, [1], 1, 1))."""



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal, Optional, Union, overload

from numpy.typing import NDArray
from typing_extensions import ParamSpec

from careamics.config.support import SupportedData
from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.file_io.read import ReadFunc

from .image_stack_loader import (
    ImageStackLoader,
    SupportedDataDev,
    get_image_stack_loader,
)

P = ParamSpec("P")


# Define overloads for each implemented ImageStackLoader case
# Array case
@overload
def create_patch_extractor(
    source: Sequence[NDArray], axes: str, data_type: Literal[SupportedData.ARRAY]
) -> PatchExtractor:
    """
    Create a patch extractor from a sequence of numpy arrays.

    Parameters
    ----------
    source: sequence of numpy.ndarray
        The source arrays of the data.
    data_config: DataConfig
        The data configuration, `data_config.data_type` should have the value "array",
        and `data_config.axes` should describe the axes of every array in the `source`.

    Returns
    -------
    PatchExtractor
    """


# TIFF and ZARR case
@overload
def create_patch_extractor(
    source: Sequence[Path],
    axes: str,
    data_type: Literal[SupportedData.TIFF, SupportedDataDev.ZARR],
) -> PatchExtractor:
    """
    Create a patch extractor from a sequence of files that match our supported types.

    Supported file types include TIFF and ZARR.

    If the files are ZARR files they must follow the OME standard. If you have ZARR
    files that do not follow the OME standard, see documentation on how to create
    a custom `image_stack_loader`. (TODO: Add link).

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    data_config: DataConfig
        The data configuration, `data_config.data_type` should have the value "tiff" or
        "zarr", and `data_config.axes` should describe the axes of every image in the
        `source`.

    Returns
    -------
    PatchExtractor
    """


# Custom file type case (loaded into memory)
@overload
def create_patch_extractor(
    source: Any,
    axes: str,
    data_type: Literal[SupportedData.CUSTOM],
    *,
    read_func: ReadFunc,
    read_kwargs: dict[str, Any],
) -> PatchExtractor:
    """
    Create a patch extractor from a sequence of files of a custom type.

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    data_config: DataConfig
        The data configuration, `data_config.data_type` should have the value "custom".
    read_func : ReadFunc
        A function to read the custom file type, see the `ReadFunc` protocol.
    read_kwargs : dict of {str: Any}
        Kwargs that will be passed to the custom `read_func`.

    Returns
    -------
    PatchExtractor
    """


# Custom ImageStackLoader case
@overload
def create_patch_extractor(
    source: Any,
    axes: str,
    data_type: Literal[SupportedData.CUSTOM],
    image_stack_loader: ImageStackLoader[P],
    *args: P.args,
    **kwargs: P.kwargs,
) -> PatchExtractor:
    """
    Create a patch extractor using a custom `ImageStackLoader`.

    The custom image stack loader must follow the `ImageStackLoader` protocol, i.e.
    it must have the following function signature:
    ```
    def image_loader_example(
        source: Any, data_config: DataConfig, *args, **kwargs
    ) -> Sequence[ImageStack]:
    ```

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    data_config: DataConfig
        The data configuration, `data_config.data_type` should have the value "custom".
    image_stack_loader: ImageStackLoader
        A custom image stack loader callable.
    *args: Any
        Positional arguments that will be passed to the custom image stack loader.
    **kwargs: Any
        Keyword arguments that will be passed to the custom image stack loader.

    Returns
    -------
    PatchExtractor
    """


# final overload to match the implentation function signature
# Need this so it works later in the code
#   (bec there aren't created overloads for create_patch_extractors below)
@overload
def create_patch_extractor(
    source: Any,
    axes: str,
    data_type: Union[SupportedData, SupportedDataDev],
    image_stack_loader: Optional[ImageStackLoader[P]] = None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> PatchExtractor: ...


def create_patch_extractor(
    source: Any,
    axes: str,
    data_type: Union[SupportedData, SupportedDataDev],
    image_stack_loader: Optional[ImageStackLoader[P]] = None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> PatchExtractor:
    # TODO: Do we need to catch data_config.data_type and source mismatches?
    #   e.g. data_config.data_type is "array" but source is not Sequence[NDArray]
    loader: ImageStackLoader[P] = get_image_stack_loader(data_type, image_stack_loader)
    image_stacks = loader(source, axes, *args, **kwargs)
    return PatchExtractor(image_stacks)


# TODO: Remove this and just call `create_patch_extractor` within the Dataset class
# Keeping for consistency for now
def create_patch_extractors(
    source: Any,
    target_source: Optional[Any],
    axes: str,
    data_type: Union[SupportedData, SupportedDataDev],
    image_stack_loader: Optional[ImageStackLoader] = None,
    *args,
    **kwargs,
) -> tuple[PatchExtractor, Optional[PatchExtractor]]:

    # --- data extractor
    patch_extractor: PatchExtractor = create_patch_extractor(
        source,
        axes,
        data_type,
        image_stack_loader,
        *args,
        **kwargs,
    )
    # --- optional target extractor
    if target_source is not None:
        target_patch_extractor = create_patch_extractor(
            target_source,
            axes,
            data_type,
            image_stack_loader,
            *args,
            **kwargs,
        )

        return patch_extractor, target_patch_extractor

    return patch_extractor, None



"""CAREamics Pydantic configuration models.

To maintain clarity at the module level, we follow the following naming conventions:
`*_model` is specific for sub-configurations (e.g. architecture, data, algorithm),
while `*_configuration` is reserved for the main configuration models, including the
`Configuration` base class and its algorithm-specific child classes.
"""

__all__ = [
    "CAREAlgorithm",
    "CheckpointModel",
    "Configuration",
    "DataConfig",
    "GaussianMixtureNMConfig",
    "InferenceConfig",
    "LVAELossConfig",
    "MultiChannelNMConfig",
    "N2NAlgorithm",
    "N2VAlgorithm",
    "TrainingConfig",
    "UNetBasedAlgorithm",
    "VAEBasedAlgorithm",
    "algorithm_factory",
    "create_care_configuration",
    "create_n2n_configuration",
    "create_n2v_configuration",
    "load_configuration",
    "save_configuration",
]

from .algorithms import (
    CAREAlgorithm,
    N2NAlgorithm,
    N2VAlgorithm,
    UNetBasedAlgorithm,
    VAEBasedAlgorithm,
)
from .callback_model import CheckpointModel
from .configuration import Configuration
from .configuration_factories import (
    algorithm_factory,
    create_care_configuration,
    create_n2n_configuration,
    create_n2v_configuration,
)
from .configuration_io import load_configuration, save_configuration
from .data import DataConfig
from .inference_model import InferenceConfig
from .loss_model import LVAELossConfig
from .nm_model import GaussianMixtureNMConfig, MultiChannelNMConfig
from .training_model import TrainingConfig



"""Pydantic model for the XYRandomRotate90 transform."""

from typing import Literal, Optional

from pydantic import ConfigDict, Field

from .transform_model import TransformModel


class XYRandomRotate90Model(TransformModel):
    """
    Pydantic model used to represent the XY random 90 degree rotation transformation.

    Attributes
    ----------
    name : Literal["XYRandomRotate90"]
        Name of the transformation.
    p : float
        Probability of applying the transform, by default 0.5.
    seed : Optional[int]
        Seed for the random number generator, by default None.
    """

    model_config = ConfigDict(
        validate_assignment=True,
    )

    name: Literal["XYRandomRotate90"] = "XYRandomRotate90"
    p: float = Field(
        0.5,
        description="Probability of applying the transform.",
        ge=0,
        le=1,
    )
    seed: Optional[int] = None



"""Autocorrelation function."""

import numpy as np
from numpy.typing import NDArray


def autocorrelation(image: NDArray) -> NDArray:
    """Compute the autocorrelation of an image.

    This method is used to explore spatial correlations in images,
    in particular in the noise.

    The autocorrelation is normalized to the zero-shift value, which is centered in
    the resulting images.

    Parameters
    ----------
    image : NDArray
        Input image.

    Returns
    -------
    numpy.ndarray
        Autocorrelation of the input image.
    """
    # normalize image
    image = (image - np.mean(image)) / np.std(image)

    # compute autocorrelation in fourier space
    image = np.fft.fftn(image)
    image = np.abs(image) ** 2
    image = np.fft.ifftn(image).real

    # normalize to zero shift value
    image = image / image.flat[0]

    # shift zero frequency to center
    image = np.fft.fftshift(image)

    return image



from pathlib import Path

from careamics.utils import cwd


def test_cwd(tmp_path: Path):
    """Test cwd context manager."""
    current_path = Path(".").absolute()
    assert current_path != tmp_path.absolute()

    # change context
    with cwd(tmp_path):
        assert Path(".").absolute() == tmp_path.absolute()

        # create a file
        file_path = Path("test.txt")
        file_path.touch()
        assert Path(tmp_path, "test.txt").absolute().exists()

    # check that we are back to the original context
    assert Path(".").absolute() == current_path



"""Losses module."""

__all__ = [
    "denoisplit_loss",
    "denoisplit_musplit_loss",
    "loss_factory",
    "mae_loss",
    "mse_loss",
    "musplit_loss",
    "n2v_loss",
]

from .fcn.losses import mae_loss, mse_loss, n2v_loss
from .loss_factory import loss_factory
from .lvae.losses import denoisplit_loss, denoisplit_musplit_loss, musplit_loss



from enum import Enum


class DataType(Enum):
    Elisa3DData = 0
    HTLIF24Data = 1
    Pavia3SeqData = 2
    TavernaSox2GolgiV2 = 3
    Dao3ChannelWithInput = 4
    ExpMicroscopyV1 = 5
    ExpMicroscopyV2 = 6
    Dao3Channel = 7
    TavernaSox2Golgi = 8
    HTIba1Ki67 = 9
    OptiMEM100_014 = 10
    SeparateTiffData = 11
    BioSR_MRC = 12
    PunctaRemoval = 13  # for the case when we have a set of differently sized crops for each channel.
    Care3D = 14


class DataSplitType(Enum):
    All = 0
    Train = 1
    Val = 2
    Test = 3


class TilingMode(Enum):
    TrimBoundary = 0
    PadBoundary = 1
    ShiftBoundary = 2



import numpy as np
import pytest

from careamics.config.tile_information import TileInformation
from careamics.dataset.tiling.tiled_patching import (
    _compute_crop_and_stitch_coords_1d,
    extract_tiles,
)


def check_extract_tiles(array: np.ndarray, tile_size, overlaps):
    """Test extracting patches randomly."""
    tile_data_generator = extract_tiles(array, tile_size, overlaps)

    tiles = []
    all_overlap_crop_coords = []
    all_stitch_coords = []

    # Assemble all tiles and their respective coordinates
    for tile_data in tile_data_generator:
        tile = tile_data[0]

        tile_info: TileInformation = tile_data[1]
        overlap_crop_coords = tile_info.overlap_crop_coords
        stitch_coords = tile_info.stitch_coords

        # add data to lists
        tiles.append(tile)
        all_overlap_crop_coords.append(overlap_crop_coords)
        all_stitch_coords.append(stitch_coords)

        # check tile shape, ignore sample dimension
        assert tile.shape[1:] == tile_size
        assert len(overlap_crop_coords) == len(stitch_coords) == len(tile_size)

    # check that each tile has a unique set of coordinates
    assert len(tiles) == len(all_overlap_crop_coords) == len(all_stitch_coords)

    # check that all values are covered by the tiles
    n_max = np.prod(array.shape)  # maximum value in the array
    unique = np.unique(np.array(tiles))  # unique values in the patches
    assert len(unique) >= n_max


@pytest.mark.parametrize(
    "tile_size, overlaps",
    [
        ((4, 4), (2, 2)),
        ((8, 8), (4, 4)),
    ],
)
def test_extract_tiles_2d(array_2D, tile_size, overlaps):
    """Test extracting tiles for prediction in 2D."""
    check_extract_tiles(array_2D, tile_size, overlaps)


@pytest.mark.parametrize(
    "tile_size, overlaps",
    [
        ((4, 4, 4), (2, 2, 2)),
        ((8, 8, 8), (4, 4, 4)),
    ],
)
def test_extract_tiles_3d(array_3D, tile_size, overlaps):
    """Test extracting tiles for prediction in 3D.

    The 3D array is a fixture of shape (1, 8, 16, 16)."""
    check_extract_tiles(array_3D, tile_size, overlaps)


@pytest.mark.parametrize("axis_size", [32, 35, 40])
@pytest.mark.parametrize("patch_size, overlap", [(16, 4), (8, 6), (16, 8), (32, 24)])
def test_compute_crop_and_stitch_coords_1d(axis_size, patch_size, overlap):
    (
        crop_coords,
        stitch_coords,
        overlap_crop_coords,
    ) = _compute_crop_and_stitch_coords_1d(axis_size, patch_size, overlap)

    # check that the number of patches is sufficient to cover the whole axis and that
    # the number of coordinates is
    # the same for all three coordinate groups
    num_patches = np.ceil((axis_size - overlap) / (patch_size - overlap)).astype(int)
    assert (
        len(crop_coords)
        == len(stitch_coords)
        == len(overlap_crop_coords)
        == num_patches
    )
    # check if 0 is the first coordinate, axis_size is last coordinate in all three
    # coordinate groups
    assert all(
        all((group[0][0] == 0, group[-1][1] == axis_size))
        for group in [crop_coords, stitch_coords]
    )
    # check if neighboring stitch coordinates are equal
    assert all(
        stitch_coords[i][1] == stitch_coords[i + 1][0]
        for i in range(len(stitch_coords) - 1)
    )

    # check that the crop coordinates cover the whole axis
    assert (
        np.sum(np.array(crop_coords)[:, 1] - np.array(crop_coords)[:, 0])
        == patch_size * num_patches
    )

    # check that the overlap crop coordinates cover the whole axis
    assert (
        np.sum(
            np.array(overlap_crop_coords)[:, 1] - np.array(overlap_crop_coords)[:, 0]
        )
        == axis_size
    )

    # check that shape of all cropped tiles is equal
    assert np.array_equal(
        np.array(overlap_crop_coords)[:, 1] - np.array(overlap_crop_coords)[:, 0],
        np.array(stitch_coords)[:, 1] - np.array(stitch_coords)[:, 0],
    )



import itertools
from collections.abc import Sequence
from typing import Optional

import numpy as np
from typing_extensions import ParamSpec

from .patching_strategy_protocol import PatchSpecs

P = ParamSpec("P")


# TODO: this is an unfinished prototype based on current tiling implementation
#  not guaranteed to work!
class SequentialPatchingStrategy:
    # TODO: docs
    def __init__(
        self,
        data_shapes: Sequence[Sequence[int]],
        patch_size: Sequence[int],
        overlap: Optional[Sequence[int]] = None,
    ):
        self.data_shapes = data_shapes
        self.patch_size = patch_size
        if overlap is None:
            overlap = [0] * len(patch_size)
        self.overlap = np.asarray(overlap)

        self.patch_specs: list[PatchSpecs] = self._initialize_patch_specs()

    @property
    def n_patches(self) -> int:
        return len(self.patch_specs)

    def get_patch_spec(self, index: int) -> PatchSpecs:
        return self.patch_specs[index]

    def _compute_coords_1d(
        self, patch_size: int, spatial_shape: int, overlap: int
    ) -> list[tuple[int, int]]:
        step = patch_size - overlap
        crop_coords = []

        current_pos = 0
        while current_pos <= spatial_shape - patch_size:
            crop_coords.append((current_pos, current_pos + patch_size))
            current_pos += step

        if crop_coords[-1][1] < spatial_shape:
            crop_coords.append((spatial_shape - patch_size, spatial_shape))

        return crop_coords

    def _initialize_patch_specs(self) -> list[PatchSpecs]:
        patch_specs: list[PatchSpecs] = []
        for data_idx, data_shape in enumerate(self.data_shapes):

            data_spatial_shape = data_shape[-len(self.patch_size) :]
            coords_list = [
                self._compute_coords_1d(
                    self.patch_size[i], data_spatial_shape[i], self.overlap[i]
                )
                for i in range(len(self.patch_size))
            ]
            for sample_idx in range(data_shape[0]):
                for crop_coord in itertools.product(*coords_list):
                    patch_specs.append(
                        PatchSpecs(
                            data_idx=data_idx,
                            sample_idx=sample_idx,
                            coords=tuple(coord[0] for coord in crop_coord),
                            patch_size=self.patch_size,
                        )
                    )
        return patch_specs



# %%
import numpy as np

# %%
from careamics.config.support import SupportedData
from careamics.dataset_ng.patch_extractor import create_patch_extractor
from careamics.dataset_ng.patch_extractor.image_stack import InMemoryImageStack
from careamics.dataset_ng.patching_strategies import RandomPatchingStrategy

# %%
array = np.arange(36).reshape(6, 6)
image_stack = InMemoryImageStack.from_array(data=array, axes="YX")
image_stack.extract_patch(sample_idx=0, coords=(2, 2), patch_size=(3, 3))

# %%
rng = np.random.default_rng()

# %%
# define example data
array1 = np.arange(36).reshape(1, 6, 6)
array2 = np.arange(50).reshape(2, 5, 5)
target1 = rng.integers(0, 1, size=array1.shape, endpoint=True)
target2 = rng.integers(0, 1, size=array2.shape, endpoint=True)

# %%
print(array1)
print(array2)
print(target1)
print(target2)

# %%
# define example readers
input_patch_extractor = create_patch_extractor(
    [array1, array2], axes="SYX", data_type=SupportedData.ARRAY
)
target_patch_extractor = create_patch_extractor(
    [target1, target2], axes="SYX", data_type=SupportedData.ARRAY
)

# %%
# generate random patch specification
data_shapes = [
    image_stack.data_shape for image_stack in input_patch_extractor.image_stacks
]
patch_specs_generator = RandomPatchingStrategy(data_shapes, patch_size=(2, 2))
patch_specs = patch_specs_generator.get_patch_spec(18)

# %%
# extract a subset of patches
input_patch_extractor.extract_patch(**patch_specs)

# %%
target_patch_extractor.extract_patch(**patch_specs)



from collections.abc import Sequence
from pathlib import Path
from typing import (
    Any,
    Optional,
    Protocol,
    Union,
)

from numpy.typing import NDArray
from typing_extensions import ParamSpec

from careamics.config.support import SupportedData
from careamics.file_io.read import ReadFunc
from careamics.utils import BaseEnum

from .image_stack import ImageStack, InMemoryImageStack, ZarrImageStack

P = ParamSpec("P")


class SupportedDataDev(str, BaseEnum):
    ZARR = "zarr"


class ImageStackLoader(Protocol[P]):
    """
    Protocol to define how `ImageStacks` should be loaded.

    An `ImageStackLoader` is a callable that must take the `source` of the data as the
    first argument, and the data `axes` as the second argument.

    Additional `*args` and `**kwargs` are allowed, but they should only be used to
    determine _how_ the data is loaded, not _what_ data is loaded. The `source`
    argument has to wholly determine _what_ data is loaded, this is because,
    downstream, both an input-source and a target-source have to be specified but they
    will share `*args` and `**kwargs`.

    An `ImageStackLoader` must return a sequence of the `ImageStack` class. This could
    be a sequence of one of the existing concrete implementations, such as
    `ZarrImageStack`, or a custom user defined `ImageStack`.

    Example
    -------
    The following example demonstrates how an `ImageStackLoader` could be defined
    for loading non-OME Zarr images. Returning a list of `ZarrImageStack` instances.

    >>> from typing import TypedDict

    >>> from zarr.storage import FSStore

    >>> from careamics.config import DataConfig
    >>> from careamics.dataset_ng.patch_extractor.image_stack import ZarrImageStack

    >>> # Define a zarr source
    >>> # It encompasses multiple arguments that determine what data will be loaded
    >>> class ZarrSource(TypedDict):
    ...     store: FSStore
    ...     data_paths: Sequence[str]

    >>> def custom_image_stack_loader(
    ...     source: ZarrSource, axes: str, *args, **kwargs
    ... ) -> list[ZarrImageStack]:
    ...     image_stacks = [
    ...         ZarrImageStack(store=source["store"], data_path=data_path, axes=axes)
    ...         for data_path in source["data_paths"]
    ...     ]
    ...     return image_stacks

    TODO: show example use in the `CAREamicsDataset`

    The example above defines a `ZarrSource` dict because to determine _which_ ZARR
    images will be loaded both a ZARR store and the internal data paths need to be
    specified.
    """

    def __call__(
        self, source: Any, axes: str, *args: P.args, **kwargs: P.kwargs
    ) -> Sequence[ImageStack]: ...


def from_arrays(
    source: Sequence[NDArray], axes: str, *args, **kwargs
) -> list[InMemoryImageStack]:
    return [InMemoryImageStack.from_array(data=array, axes=axes) for array in source]


# TODO: change source to directory path? Like in current implementation
#   Advantage of having a list is the user can match input and target order themselves
def from_tiff_files(
    source: Sequence[Path], axes: str, *args, **kwargs
) -> list[InMemoryImageStack]:
    return [InMemoryImageStack.from_tiff(path=path, axes=axes) for path in source]


# TODO: change source to directory path? Like in current implementation
#   Advantage of having a list is the user can match input and target order themselves
def from_custom_file_type(
    source: Sequence[Path],
    axes: str,
    read_func: ReadFunc,
    read_kwargs: dict[str, Any],
    *args,
    **kwargs,
) -> list[InMemoryImageStack]:
    return [
        InMemoryImageStack.from_custom_file_type(
            path=path,
            axes=axes,
            read_func=read_func,
            **read_kwargs,
        )
        for path in source
    ]


def from_ome_zarr_files(
    source: Sequence[Path], axes: str, *args, **kwargs
) -> list[ZarrImageStack]:
    # NOTE: axes is unused here, in from_ome_zarr the axes are automatically retrieved
    return [ZarrImageStack.from_ome_zarr(path) for path in source]


def get_image_stack_loader(
    data_type: Union[SupportedData, SupportedDataDev],
    image_stack_loader: Optional[ImageStackLoader] = None,
) -> ImageStackLoader:
    if data_type == SupportedData.ARRAY:
        return from_arrays
    elif data_type == SupportedData.TIFF:
        return from_tiff_files
    elif data_type == "zarr":  # temp for testing until zarr is added to SupportedData
        return from_ome_zarr_files
    elif data_type == SupportedData.CUSTOM:
        if image_stack_loader is None:
            return from_custom_file_type
        else:
            return image_stack_loader
    else:
        raise ValueError



"""
This script is meant to load data, initialize the model, and provide the logic for training it.
"""

import glob
import os
import socket
import sys
from typing import Dict

import pytorch_lightning as pl
import torch
from absl import app, flags
from ml_collections.config_flags import config_flags
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.loggers import WandbLogger
from torch.utils.data import DataLoader

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
print(sys.path)

from careamics.lvae_training.dataset.data_modules import (
    LCMultiChDloader,
    MultiChDloader,
)
from careamics.lvae_training.dataset.utils.data_utils import DataSplitType
from careamics.lvae_training.lightning_module import LadderVAELight
from careamics.lvae_training.train_utils import *

FLAGS = flags.FLAGS

config_flags.DEFINE_config_file(
    "config", None, "Training configuration.", lock_config=False
)
flags.DEFINE_string("workdir", None, "Work directory.")
flags.DEFINE_enum("mode", None, ["train", "eval"], "Running mode: train or eval")
flags.DEFINE_string(
    "logdir", "/group/jug/federico/wandb_backup/", "The folder name for storing logging"
)
flags.DEFINE_string(
    "datadir", "/group/jug/federico/careamics_training/data/BioSR", "Data directory."
)
flags.DEFINE_boolean("use_max_version", False, "Overwrite the max version of the model")
flags.DEFINE_string(
    "load_ckptfpath",
    "",
    "The path to a previous ckpt from which the weights should be loaded",
)
flags.mark_flags_as_required(["workdir", "config", "mode"])


def create_dataset(
    config,
    datadir,
    eval_datasplit_type=DataSplitType.Val,
    raw_data_dict=None,
    skip_train_dataset=False,
    kwargs_dict=None,
):

    if kwargs_dict is None:
        kwargs_dict = {}

    datapath = datadir

    # Hard-coded parameters (used to be in the config file)
    normalized_input = True
    use_one_mu_std = True
    train_aug_rotate = False
    enable_random_cropping = True
    lowres_supervision = False

    # 1) Data loader for Lateral Contextualization
    if (
        "multiscale_lowres_count" in config.data
        and config.data.multiscale_lowres_count is not None
    ):
        # Get padding attributes
        if "padding_kwargs" not in kwargs_dict:
            padding_kwargs = {}
            if "padding_mode" in config.data and config.data.padding_mode is not None:
                padding_kwargs["mode"] = config.data.padding_mode
            else:
                padding_kwargs["mode"] = "reflect"
            if "padding_value" in config.data and config.data.padding_value is not None:
                padding_kwargs["constant_values"] = config.data.padding_value
            else:
                padding_kwargs["constant_values"] = None
        else:
            padding_kwargs = kwargs_dict.pop("padding_kwargs")

        train_data = (
            None
            if skip_train_dataset
            else LCMultiChDloader(
                config.data,
                datapath,
                datasplit_type=DataSplitType.Train,
                val_fraction=0.1,
                test_fraction=0.1,
                normalized_input=normalized_input,
                use_one_mu_std=use_one_mu_std,
                enable_rotation_aug=train_aug_rotate,
                enable_random_cropping=enable_random_cropping,
                num_scales=config.data.multiscale_lowres_count,
                lowres_supervision=lowres_supervision,
                padding_kwargs=padding_kwargs,
                **kwargs_dict,
                allow_generation=True,
            )
        )
        max_val = train_data.get_max_val()

        val_data = LCMultiChDloader(
            config.data,
            datapath,
            datasplit_type=eval_datasplit_type,
            val_fraction=0.1,
            test_fraction=0.1,
            normalized_input=normalized_input,
            use_one_mu_std=use_one_mu_std,
            enable_rotation_aug=False,  # No rotation aug on validation
            enable_random_cropping=False,
            # No random cropping on validation. Validation is evaluated on determistic grids
            num_scales=config.data.multiscale_lowres_count,
            lowres_supervision=lowres_supervision,
            padding_kwargs=padding_kwargs,
            allow_generation=False,
            **kwargs_dict,
            max_val=max_val,
        )
    # 2) Vanilla data loader
    else:
        train_data_kwargs = {"allow_generation": True, **kwargs_dict}
        val_data_kwargs = {"allow_generation": False, **kwargs_dict}

        train_data_kwargs["enable_random_cropping"] = enable_random_cropping
        val_data_kwargs["enable_random_cropping"] = False

        train_data = (
            None
            if skip_train_dataset
            else MultiChDloader(
                data_config=config.data,
                fpath=datapath,
                datasplit_type=DataSplitType.Train,
                val_fraction=0.1,
                test_fraction=0.1,
                normalized_input=normalized_input,
                use_one_mu_std=use_one_mu_std,
                enable_rotation_aug=train_aug_rotate,
                **train_data_kwargs,
            )
        )

        max_val = train_data.get_max_val()
        val_data = MultiChDloader(
            data_config=config.data,
            fpath=datapath,
            datasplit_type=eval_datasplit_type,
            val_fraction=0.1,
            test_fraction=0.1,
            normalized_input=normalized_input,
            use_one_mu_std=use_one_mu_std,
            enable_rotation_aug=False,  # No rotation aug on validation
            max_val=max_val,
            **val_data_kwargs,
        )

    # For normalizing, we should be using the training data's mean and std.
    mean_val, std_val = train_data.compute_mean_std()
    train_data.set_mean_std(mean_val, std_val)
    val_data.set_mean_std(mean_val, std_val)

    return train_data, val_data


def create_model_and_train(
    config: ml_collections.ConfigDict,
    data_mean: Dict[str, torch.Tensor],
    data_std: Dict[str, torch.Tensor],
    logger: WandbLogger,
    checkpoint_callback: ModelCheckpoint,
    train_loader: DataLoader,
    val_loader: DataLoader,
):
    # tensorboard previous files.
    for filename in glob.glob(config.workdir + "/events*"):
        os.remove(filename)

    # checkpoints
    for filename in glob.glob(config.workdir + "/*.ckpt"):
        os.remove(filename)

    if "num_targets" in config.model:
        target_ch = config.model.num_targets
    else:
        target_ch = config.data.get("num_channels", 2)

    # Instantiate the model (lightning wrapper)
    model = LadderVAELight(
        data_mean=data_mean, data_std=data_std, config=config, target_ch=target_ch
    )

    # Load pre-trained weights if any
    if config.training.pre_trained_ckpt_fpath:
        print("Starting with pre-trained model", config.training.pre_trained_ckpt_fpath)
        checkpoint = torch.load(config.training.pre_trained_ckpt_fpath)
        _ = model.load_state_dict(checkpoint["state_dict"], strict=False)

    estop_monitor = config.model.get("monitor", "val_loss")
    estop_mode = MetricMonitor(estop_monitor).mode()

    callbacks = [
        EarlyStopping(
            monitor=estop_monitor,
            min_delta=1e-6,
            patience=config.training.earlystop_patience,
            verbose=True,
            mode=estop_mode,
        ),
        checkpoint_callback,
        LearningRateMonitor(logging_interval="epoch"),
    ]

    logger.experiment.config.update(config.to_dict())
    # wandb.init(config=config)
    trainer = pl.Trainer(
        accelerator="gpu",
        max_epochs=config.training.max_epochs,
        gradient_clip_val=config.training.grad_clip_norm_value,
        gradient_clip_algorithm=config.training.gradient_clip_algorithm,
        logger=logger,
        callbacks=callbacks,
        # limit_train_batches = config.training.limit_train_batches,
        precision=config.training.precision,
    )
    trainer.fit(model, train_loader, val_loader)


def train_network(
    train_loader: DataLoader,
    val_loader: DataLoader,
    data_mean: Dict[str, torch.Tensor],
    data_std: Dict[str, torch.Tensor],
    config: ml_collections.ConfigDict,
    model_name: str,
    logdir: str,
):
    ckpt_monitor = config.model.get("monitor", "val_loss")
    ckpt_mode = MetricMonitor(ckpt_monitor).mode()
    checkpoint_callback = ModelCheckpoint(
        monitor=ckpt_monitor,
        dirpath=config.workdir,
        filename=model_name + "_best",
        save_last=True,
        save_top_k=1,
        mode=ckpt_mode,
    )
    checkpoint_callback.CHECKPOINT_NAME_LAST = model_name + "_last"
    logger = WandbLogger(
        name=os.path.join(config.hostname, config.exptname),
        save_dir=logdir,
        project="Disentanglement",
    )

    create_model_and_train(
        config=config,
        data_mean=data_mean,
        data_std=data_std,
        logger=logger,
        checkpoint_callback=checkpoint_callback,
        train_loader=train_loader,
        val_loader=val_loader,
    )


def main(argv):
    config = FLAGS.config

    assert os.path.exists(FLAGS.workdir)
    cur_workdir, relative_path = get_workdir(
        config, FLAGS.workdir, FLAGS.use_max_version
    )
    print(f"Saving training to {cur_workdir}")

    config.workdir = cur_workdir
    config.exptname = relative_path
    config.hostname = socket.gethostname()
    config.datadir = FLAGS.datadir
    config.training.pre_trained_ckpt_fpath = FLAGS.load_ckptfpath

    if FLAGS.mode == "train":
        set_logger(workdir=cur_workdir)
        raw_data_dict = None

        # From now on, config cannot be changed.
        config = ml_collections.FrozenConfigDict(config)
        log_config(config, cur_workdir)

        train_data, val_data = create_dataset(
            config, FLAGS.datadir, raw_data_dict=raw_data_dict
        )

        mean_dict, std_dict = get_mean_std_dict_for_model(config, train_data)

        batch_size = config.training.batch_size
        shuffle = True
        train_dloader = DataLoader(
            train_data,
            pin_memory=False,
            num_workers=config.training.num_workers,
            shuffle=shuffle,
            batch_size=batch_size,
        )
        val_dloader = DataLoader(
            val_data,
            pin_memory=False,
            num_workers=config.training.num_workers,
            shuffle=False,
            batch_size=batch_size,
        )

        train_network(
            train_loader=train_dloader,
            val_loader=val_dloader,
            data_mean=mean_dict,
            data_std=std_dict,
            config=config,
            model_name="BaselineVAECL",
            logdir=FLAGS.logdir,
        )

    elif FLAGS.mode == "eval":
        pass
    else:
        raise ValueError(f"Mode {FLAGS.mode} not recognized.")


if __name__ == "__main__":
    app.run(main)



from careamics.config.architectures import ArchitectureModel


def test_model_dump():
    """Test that architecture keyword is removed from the model dump."""
    model_params = {"architecture": "LeCorbusier"}
    model = ArchitectureModel(**model_params)

    # dump model
    model_dict = model.model_dump()
    assert model_dict == {}



import pytest
import torch

from careamics.models.layers import MaxBlurPool
from careamics.models.unet import UNet


@pytest.mark.parametrize("depth", [1, 3, 5])
def test_unet_depth(depth):
    """Test that the UNet has the correct number of down and up convs
    with respect to the depth."""
    model = UNet(conv_dims=2, depth=depth)

    # check that encoder has the right number of down convs
    counter_down = 0
    for _, layer in model.encoder.encoder_blocks.named_children():
        if type(layer).__name__ == "Conv_Block":
            counter_down += 1

    assert counter_down == depth

    # check that decoder has the right number of up convs
    counter_up = 0
    for _, layer in model.decoder.decoder_blocks.named_children():
        if type(layer).__name__ == "Conv_Block":
            counter_up += 1

    assert counter_up == depth


@pytest.mark.parametrize(
    "input_shape",
    [
        (1, 1, 1024, 1024),
        (1, 1, 512, 512),
        (1, 1, 256, 256),
        (1, 1, 128, 128),
        (1, 1, 64, 64),
        (1, 1, 32, 32),
        (1, 1, 16, 16),
        (1, 1, 8, 8),
    ],
)
def test_blurpool2d(input_shape):
    """Test that the BlurPool2d layer works as expected."""
    layer = MaxBlurPool(dim=2, kernel_size=3)
    assert layer(torch.randn(input_shape)).shape == tuple(
        [1, 1] + [i // 2 for i in input_shape[2:]]
    )


@pytest.mark.parametrize(
    "input_shape",
    [
        (1, 1, 256, 256, 256),
        (1, 1, 128, 128, 128),
        (1, 1, 64, 128, 128),
        (1, 1, 64, 64, 64),
        (1, 1, 32, 64, 64),
        (1, 1, 32, 32, 32),
        (1, 1, 16, 16, 16),
        (1, 1, 8, 8, 8),
    ],
)
def test_blurpool3d(input_shape):
    """Test that the BlurPool3d layer works as expected."""
    layer = MaxBlurPool(dim=3, kernel_size=3)
    assert layer(torch.randn(input_shape)).shape == tuple(
        [1, 1] + [i // 2 for i in input_shape[2:]]
    )


@pytest.mark.parametrize("independent_channels", [True, False])
def test_independent_channels(independent_channels):
    """
    Test that with independent channels there is no connections between the
    parallel networks.
    """

    n_channels = 2
    input_shape = (1, n_channels, 8, 8)

    # input 1
    x1 = torch.randn(input_shape)
    # input 2: channel 1 same as input 1, channel 2 all zeros
    x2 = x1.clone()
    x2[:, 1] = 0

    model = UNet(
        conv_dims=2,
        in_channels=n_channels,
        num_classes=n_channels,
        independent_channels=independent_channels,
    )
    model.eval()

    y1 = model(x1)
    y2 = model(x2)

    # channel 2 different between inputs => channel 2 different between outputs
    assert (y1[:, 1] != y2[:, 1]).any()

    if independent_channels:
        # when channel independent
        # channel 1 same between inputs => channel 1 same between outputs
        assert (y1[:, 0] == y2[:, 0]).all()
    else:
        # when not independent
        # channel 2 different between inputs => all channels different between outputs
        assert (y1[:, 0] != y2[:, 0]).any()



import numpy as np
import pytest

from careamics.dataset_ng.patching_strategies.random_patching import (
    RandomPatchingStrategy,
    _calc_n_patches,
    _generate_random_coords,
)


# Note: get_patch_spec for RandomPatchingStrategy and FixedRandomStrategy is already
#   tested in test_all_strategies
@pytest.mark.parametrize(
    "data_shapes,patch_size,expected_patches",
    [
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 8), 49],
        [[(2, 1, 32, 32), (1, 1, 19, 37), (3, 1, 14, 9)], (8, 5), 82],
        [[(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)], (8, 8, 8), 169],
        [[(2, 1, 32, 32, 32), (1, 1, 19, 37, 23), (3, 1, 14, 9, 12)], (8, 5, 7), 312],
    ],
)
def test_calc_patch_bins(data_shapes, patch_size, expected_patches):
    """Test bins are created as expected"""
    image_stack_index_bins, sample_index_bins, sample_bins = (
        RandomPatchingStrategy._calc_bins(data_shapes, patch_size)
    )
    assert image_stack_index_bins[-1] == sample_index_bins[-1] == expected_patches

    # create image_stack_bins from sample_index_bins and sample_bins
    # The idea to find the bin boundaries in sample_index_bins
    #   that are aligned to the image_stack_bins
    new_image_stack_bins = []
    for bin_boundary in sample_bins:
        idx = bin_boundary - 1
        new_image_stack_bins.append(sample_index_bins[idx])

    assert (np.array(image_stack_index_bins) == new_image_stack_bins).all()


@pytest.mark.parametrize(
    "data_shape,patch_size,expected_patches",
    [
        [(1, 1, 19, 37), (8, 8), 11],
        [(1, 1, 19, 37), (8, 5), 18],
        [(1, 1, 19, 37, 23), (8, 8, 8), 32],
        [(1, 1, 19, 37, 23), (8, 5, 7), 58],
    ],
)
def test_n_patches(data_shape, patch_size, expected_patches):
    spatial_shape = data_shape[2:]
    n_patches = _calc_n_patches(spatial_shape, patch_size)

    assert n_patches == expected_patches


def test_n_patches_raises():
    spatial_shape = (8, 8)
    patch_size = (2, 2, 2)
    with pytest.raises(ValueError):
        _calc_n_patches(spatial_shape, patch_size)


@pytest.mark.parametrize(
    "data_shape,patch_size,iterations",
    [
        [(1, 1, 19, 37), (8, 8), 11],
        [(1, 1, 19, 37), (8, 5), 18],
        [(1, 1, 19, 37, 23), (8, 8, 8), 32],
        [(1, 1, 19, 37, 23), (8, 5, 7), 58],
    ],
)
def test_random_coords(data_shape, patch_size, iterations):
    spatial_shape = data_shape[2:]
    rng = np.random.default_rng(42)
    for _ in range(iterations):
        coords = np.array(_generate_random_coords(spatial_shape, patch_size, rng))
        # validate patch is within spatial bounds
        assert (0 <= coords).all()
        assert (coords + patch_size < np.array(spatial_shape)).all()


def test_random_coords_raises():
    spatial_shape = (8, 8)
    patch_size = (2, 2, 2)
    rng = np.random.default_rng(42)
    with pytest.raises(ValueError):
        _generate_random_coords(spatial_shape, patch_size, rng)



import pytest

from careamics.config.training_model import TrainingConfig


def test_training_wrong_values_by_assignments(minimum_training: dict):
    """Test that wrong values cause an error during assignment."""
    training = TrainingConfig(**minimum_training)

    # num_epochs
    training.num_epochs = 2
    with pytest.raises(ValueError):
        training.num_epochs = -1

    training.max_steps = 2
    with pytest.raises(ValueError):
        training.max_steps = 0

    training.precision = "16-mixed"
    with pytest.raises(ValueError):
        training.precision = "16"



import pytest
import torch
from onnx import checker

from careamics.config import UNetBasedAlgorithm
from careamics.lightning.lightning_module import FCNModule


# TODO: move a module for special integration tests
# TODO revisit for specific algorithm configuration
@pytest.mark.parametrize(
    "algorithm, architecture, conv_dim, n2v2, loss, shape",
    [
        ("n2n", "UNet", 2, False, "mae", (16, 16)),  # n2n 2D model
        ("n2n", "UNet", 3, False, "mae", (8, 16, 16)),  # n2n 3D model
        ("n2v", "UNet", 2, False, "n2v", (16, 16)),  # n2v 2D model
        ("n2v", "UNet", 3, False, "n2v", (8, 16, 16)),  # n2v 3D model
        ("n2v", "UNet", 2, True, "n2v", (16, 16)),  # n2v2 2D model
        ("n2v", "UNet", 3, True, "n2v", (8, 16, 16)),  # n2v2 3D model
    ],
)
def test_onnx_export(tmp_path, algorithm, architecture, conv_dim, n2v2, loss, shape):
    """Test model exportability to ONNX."""

    algo_config = {
        "algorithm": algorithm,
        "model": {
            "architecture": architecture,
            "conv_dims": conv_dim,
            "in_channels": 1,
            "num_classes": 1,
            "depth": 3,
            "n2v2": n2v2,
        },
        "loss": loss,
    }
    algo_config = UNetBasedAlgorithm(**algo_config)

    # instantiate CAREamicsKiln
    model = FCNModule(algo_config)
    # set model to evaluation mode to avoid batch dimension error
    model.model.eval()
    # create a sample input of BC(Z)XY
    x = torch.rand((1, 1, *shape))

    # create dynamic axes from the shape of the x
    dynamic_axes = {"input": {}, "output": {}}
    for i in range(len(x.shape)):
        dynamic_axes["input"][i] = f"dim_{i}"
        dynamic_axes["output"][i] = f"dim_{i}"

    torch.onnx.export(
        model,
        x,
        f"{tmp_path}/test_model.onnx",
        input_names=["input"],  # the model's input names
        output_names=["output"],  # the model's output names
        dynamic_axes=dynamic_axes,  # variable length axes,
    )

    checker.check_model(f"{tmp_path}/test_model.onnx")
